import logging
from contextlib import contextmanager
import cProfile
import pstats
import threading
from functools import wraps

from .smtlib import issymbolic

logger = logging.getLogger(__name__)


class DecorateAllMeta(type):
    @staticmethod
    def _if_enabled(f):
        """ decorator used to guard callbacks """

        @wraps(f)
        def g(self, *args, **kwargs):
            if self.is_enabled():
                return f(self, *args, **kwargs)

        return g

    def __new__(cls, name, bases, local):
        for attr in local:
            value = local[attr]
            if attr.endswith("_callback") and callable(value):
                local[attr] = cls._if_enabled(value)
        return type.__new__(cls, name, bases, local)


class Plugin(metaclass=DecorateAllMeta):
    __slots__ = ("manticore", "_enabled_key", "_plugin_context_name")

    def __init__(self):
        self.manticore = None
        classname = str(type(self)).split("'")[1]
        self._enabled_key = f"{classname}_enabled_{hex(hash(self))}"
        self._plugin_context_name = f"{classname}_context_{hex(hash(self))}"

    def enable(self):
        """ Enable all callbacks """
        with self.manticore.locked_context() as context:
            context[self._enabled_key] = True

    def disable(self):
        """ Disable all callbacks """
        with self.manticore.locked_context() as context:
            context[self._enabled_key] = False

    def is_enabled(self):
        """ True if callbacks are enabled """
        with self.manticore.locked_context() as context:
            return context.get(self._enabled_key, True)

    @property
    def name(self):
        return str(self.__class__)

    @contextmanager
    def locked_context(self, key=None, value_type=list):
        """
        A context manager that provides safe parallel access to the global Manticore context.
        This should be used to access the global Manticore context
        when parallel analysis is activated. Code within the `with` block is executed
        atomically, so access of shared variables should occur within.
        """
        plugin_context_name = self._plugin_context_name
        with self.manticore.locked_context(plugin_context_name, dict) as context:
            if key is None:
                yield context
            else:
                ctx = context.get(key, value_type())
                yield ctx
                context[key] = ctx

    @property
    def context(self):
        """ Convenient access to shared context """
        plugin_context_name = self._plugin_context_name
        if plugin_context_name not in self.manticore.context:
            self.manticore.context[plugin_context_name] = {}
        return self.manticore.context[plugin_context_name]

    def on_register(self):
        """ Called by parent manticore on registration """
        pass

    def on_unregister(self):
        """ Called be parent manticore on un-registration """
        pass

    def generate_testcase(self, state, testcase, message):
        """ Called so the plugin can attach some results to the testcase if the
            state needs it"""
        pass


def _dict_diff(d1, d2):
    """
    Produce a dict that includes all the keys in d2 that represent different values in d1, as well as values that
    aren't in d1.

    :param dict d1: First dict
    :param dict d2: Dict to compare with
    :rtype: dict
    """
    d = {}
    for key in set(d1).intersection(set(d2)):
        if d2[key] != d1[key]:
            d[key] = d2[key]
    for key in set(d2).difference(set(d1)):
        d[key] = d2[key]
    return d


class Tracer(Plugin):
    def did_execute_instruction_callback(self, state, pc, target_pc, instruction):
        state.context.setdefault("trace", []).append(pc)


class ExtendedTracer(Plugin):
    def __init__(self):
        """
        Record a detailed execution trace
        """
        super().__init__()
        self.last_dict = {}
        self.current_pc = None
        self.context_key = "e_trace"

    def get_trace(self, state):
        return state.context.get(self.context_key)

    def register_state_to_dict(self, cpu):
        d = {}
        for reg in cpu.canonical_registers:
            val = cpu.read_register(reg)
            d[reg] = val if not issymbolic(val) else "<sym>"
        return d

    def will_execute_instruction_callback(self, state, pc, instruction):
        self.current_pc = pc

    def did_execute_instruction_callback(self, state, pc, target_pc, instruction):
        reg_state = self.register_state_to_dict(state.cpu)
        entry = {"type": "regs", "values": _dict_diff(self.last_dict, reg_state)}
        self.last_dict = reg_state
        state.context.setdefault(self.context_key, []).append(entry)

    def will_read_memory_callback(self, state, where, size):
        if self.current_pc == where:
            return

    def did_read_memory_callback(self, state, where, value, size):
        if self.current_pc == where:
            return

    def will_write_memory_callback(self, state, where, value, size):
        if self.current_pc == where:
            return

    def did_write_memory_callback(self, state, where, value, size):
        if self.current_pc == where:
            raise Exception

        entry = {"type": "mem_write", "where": where, "value": value, "size": size}
        state.context.setdefault(self.context_key, []).append(entry)


class Follower(Plugin):
    def __init__(self, trace):
        self.index = 0
        self.trace = trace
        self.last_instruction = None
        self.symbolic_ranges = []
        self.active = True
        super().__init__()

    def add_symbolic_range(self, pc_start, pc_end):
        self.symbolic_ranges.append((pc_start, pc_end))

    def get_next(self, type):
        event = self.trace[self.index]
        assert event["type"] == type
        self.index += 1
        return event

    def did_write_memory_callback(self, state, where, value, size):
        if not self.active:
            return
        write = self.get_next("mem_write")

        if not issymbolic(value):
            return

        assert write["where"] == where and write["size"] == size
        # state.constrain(value == write['value'])

    def did_execute_instruction_callback(self, state, last_pc, pc, insn):
        if not self.active:
            return
        event = self.get_next("regs")
        self.last_instruction = event["values"]
        if issymbolic(pc):
            state.constrain(state.cpu.RIP == self.last_instruction["RIP"])
        else:
            for start, stop in self.symbolic_ranges:
                if start <= pc <= stop:
                    self.active = False


class RecordSymbolicBranches(Plugin):
    def did_execute_instruction_callback(self, state, last_pc, target_pc, instruction):
        if state.context.get("forking_pc", False):
            branches = state.context.setdefault("branches", {})
            branch = (last_pc, target_pc)
            if branch in branches:
                branches[branch] += 1
            else:
                branches[branch] = 1
            state.context["forking_pc"] = False

        if issymbolic(target_pc):
            state.context["forking_pc"] = True


class InstructionCounter(Plugin):
    def will_terminate_state_callback(self, state, ex):
        if state is None:  # FIXME Can it be None?
            return
        state_instructions_count = state.context.get("instructions_count", 0)

        with self.manticore.locked_context() as manticore_context:
            manticore_instructions_count = manticore_context.get("instructions_count", 0)
            manticore_context["instructions_count"] = (
                manticore_instructions_count + state_instructions_count
            )

    def did_execute_instruction_callback(self, state, prev_pc, target_pc, instruction):
        address = prev_pc
        if not issymbolic(address):
            count = state.context.get("instructions_count", 0)
            state.context["instructions_count"] = count + 1

    def did_run_callback(self):
        _shared_context = self.manticore.context
        instructions_count = _shared_context.get("instructions_count", 0)
        logger.info("Instructions executed: %d", instructions_count)


class Visited(Plugin):
    def __init__(self, coverage_file="visited.txt"):
        super().__init__()
        self.coverage_file = coverage_file

    def will_terminate_state_callback(self, state, ex):
        if state is None:
            return
        state_visited = state.context.get("visited_since_last_fork", set())
        with self.manticore.locked_context() as manticore_context:
            manticore_visited = manticore_context.get("visited", set())
            manticore_context["visited"] = manticore_visited.union(state_visited)

    def will_fork_state_callback(self, state, expression, values, policy):
        state_visited = state.context.get("visited_since_last_fork", set())
        with self.manticore.locked_context() as manticore_context:
            manticore_visited = manticore_context.get("visited", set())
            manticore_context["visited"] = manticore_visited.union(state_visited)
        state.context["visited_since_last_fork"] = set()

    def did_execute_instruction_callback(self, state, prev_pc, target_pc, instruction):
        state.context.setdefault("visited_since_last_fork", set()).add(prev_pc)
        state.context.setdefault("visited", set()).add(prev_pc)

    def did_run_callback(self):
        _shared_context = self.manticore.context
        executor_visited = _shared_context.get("visited", set())
        # Fixme this is duplicated?
        if self.coverage_file is not None:
            with self.manticore._output.save_stream(self.coverage_file) as f:
                for m in executor_visited:
                    f.write(f"0x{m:016x}\n")
        logger.info("Coverage: %d different instructions executed", len(executor_visited))


class Profiler(Plugin):
    data = threading.local()

    def will_start_worker_callback(self, id):
        self.data.profile = cProfile.Profile()
        self.data.profile.enable()

    def did_terminate_worker_callback(self, id):
        self.data.profile.disable()
        self.data.profile.create_stats()
        with self.manticore.locked_context("_profiling_stats", dict) as profiling_stats:
            profiling_stats[id] = self.data.profile.stats.items()

    def did_terminate_execution_callback(self, output):
        with output.save_stream("profiling.bin", binary=True) as f:
            self.save_profiling_data(f)

    def get_profiling_data(self):
        class PstatsFormatted:
            def __init__(self, d):
                self.stats = dict(d)

            def create_stats(self):
                pass

        with self.manticore.locked_context("_profiling_stats") as profiling_stats:
            ps = None
            for item in profiling_stats.values():
                try:
                    stat = PstatsFormatted(item)
                    if ps is None:
                        ps = pstats.Stats(stat)
                    else:
                        ps.add(stat)
                except TypeError:
                    logger.info("Incorrectly formatted profiling information in _stats, skipping")
        return ps

    def save_profiling_data(self, stream=None):
        """:param stream: an output stream to write the profiling data """
        ps = self.get_profiling_data()
        # XXX(yan): pstats does not support dumping to a file stream, only to a file
        # name. Below is essentially the implementation of pstats.dump_stats() without
        # the extra open().
        if stream is not None:
            import marshal

            marshal.dump(ps.stats, stream)


# TODO document all callbacks
class ExamplePlugin(Plugin):
    def will_open_transaction_callback(self, state, tx):
        logger.info("will open a transaction %r %r", state, tx)

    def will_close_transaction_callback(self, state, tx):
        logger.info("will close a transaction %r %r", state, tx)

    def will_decode_instruction_callback(self, state, pc):
        logger.info("will_decode_instruction %r %r", state, pc)

    def will_execute_instruction_callback(self, state, pc, instruction):
        logger.info("will_execute_instruction %r %r %r", state, pc, instruction)

    def did_execute_instruction_callback(self, state, pc, target_pc, instruction):
        logger.info("did_execute_instruction %r %r %r %r", state, pc, target_pc, instruction)

    def will_start_run_callback(self, state):
        """ Called once at the beginning of the run.
            state is the initial root state
        """
        logger.info("will_start_run")

    def did_run_callback(self):
        logger.info("did_run")

    def will_fork_state_callback(self, parent_state, expression, solutions, policy):
        logger.info("will_fork_state %r %r %r %r", parent_state, expression, solutions, policy)

    def did_fork_state_callback(self, child_state, expression, new_value, policy):
        logger.info("did_fork_state %r %r %r %r", child_state, expression, new_value, policy)

    def did_load_state_callback(self, state, state_id):
        logger.info("did_load_state %r %r", state, state_id)

    def did_enqueue_state_callback(self, state, state_id):
        logger.info("did_enqueue_state %r %r", state, state_id)

    def will_terminate_state_callback(self, state, exception):
        logger.info("will_terminate_state %r %r", state, exception)

    def will_generate_testcase_callback(self, state, testcase, message):
        logger.info("will_generate_testcase %r %r %r", state, testcase, message)

    def will_read_memory_callback(self, state, where, size):
        logger.info("will_read_memory %r %r %r", state, where, size)

    def did_read_memory_callback(self, state, where, value, size):
        logger.info("did_read_memory %r %r %r %r", state, where, value, size)

    def will_write_memory_callback(self, state, where, value, size):
        logger.info("will_write_memory %r %r %r", state, where, value, size)

    def did_write_memory_callback(self, state, where, value, size):
        logger.info("did_write_memory %r %r %r %r", state, where, value, size)

    def will_read_register_callback(self, state, register):
        logger.info("will_read_register %r %r", state, register)

    def did_read_register_callback(self, state, register, value):
        logger.info("did_read_register %r %r %r", state, register, value)

    def will_write_register_callback(self, state, register, value):
        logger.info("will_write_register %r %r %r", state, register, value)

    def did_write_register_callback(self, state, register, value):
        logger.info("did_write_register %r %r %r", state, register, value)



import struct
from ctypes import c_int32
from .types import (
    ConcretizeStack,
    CurGrowMemImm,
    F32,
    F32ConstImm,
    F64,
    F64ConstImm,
    GlobalVarXsImm,
    I32,
    I32ConstImm,
    I64,
    I64ConstImm,
    InvalidConversionTrap,
    LocalVarXsImm,
    MemoryImm,
    OutOfBoundsMemoryTrap,
    OverflowDivisionTrap,
    UnreachableInstructionTrap,
    Value_t,
    ZeroDivisionTrap,
)
from ..core.smtlib import Operators, BitVec, issymbolic
from ..utils.event import Eventful
from decimal import Decimal, InvalidOperation

import operator
import math

MASK_64 = (1 << 64) - 1
MASK_32 = (1 << 32) - 1


class Executor(Eventful):
    """
    Contains execution semantics for all WASM instructions that don't involve control flow (and thus only need access
    to the store and the stack).

    In lieu of annotating every single instruction with the relevant link to the docs, we direct you here:
    https://www.w3.org/TR/wasm-core-1/#a7-index-of-instructions
    """

    _published_events = {"set_global", "get_global", "set_local", "get_local"}

    def __init__(self, constraints, *args, **kwargs):

        self._mapping = {
            0x00: self.unreachable,
            0x01: self.nop,
            0x02: self.nop,  # block
            0x03: self.nop,  # loop
            0x04: self.nop,  # if
            0x05: self.nop,  # else
            0x0B: self.nop,  # end
            0x0C: self.nop,  # br
            0x0D: self.nop,  # br_if
            0x0E: self.nop,  # br_table
            0x0F: self.nop,  # return
            0x10: self.nop,  # call
            0x11: self.nop,  # call_indirect
            0x1A: self.drop,
            0x1B: self.select,
            0x20: self.get_local,
            0x21: self.set_local,
            0x22: self.tee_local,
            0x23: self.get_global,
            0x24: self.set_global,
            0x28: self.i32_load,
            0x29: self.i64_load,
            0x2A: self.f32_load,
            0x2B: self.f64_load,
            0x2C: self.i32_load8_s,
            0x2D: self.i32_load8_u,
            0x2E: self.i32_load16_s,
            0x2F: self.i32_load16_u,
            0x30: self.i64_load8_s,
            0x31: self.i64_load8_u,
            0x32: self.i64_load16_s,
            0x33: self.i64_load16_u,
            0x34: self.i64_load32_s,
            0x35: self.i64_load32_u,
            0x36: self.i32_store,
            0x37: self.i64_store,
            0x38: self.f32_store,
            0x39: self.f64_store,
            0x3A: self.i32_store8,
            0x3B: self.i32_store16,
            0x3C: self.i64_store8,
            0x3D: self.i64_store16,
            0x3E: self.i64_store32,
            0x3F: self.current_memory,
            0x40: self.grow_memory,
            0x41: self.i32_const,
            0x42: self.i64_const,
            0x43: self.f32_const,
            0x44: self.f64_const,
            0x45: self.i32_eqz,
            0x46: self.i32_eq,
            0x47: self.i32_ne,
            0x48: self.i32_lt_s,
            0x49: self.i32_lt_u,
            0x4A: self.i32_gt_s,
            0x4B: self.i32_gt_u,
            0x4C: self.i32_le_s,
            0x4D: self.i32_le_u,
            0x4E: self.i32_ge_s,
            0x4F: self.i32_ge_u,
            0x50: self.i64_eqz,
            0x51: self.i64_eq,
            0x52: self.i64_ne,
            0x53: self.i64_lt_s,
            0x54: self.i64_lt_u,
            0x55: self.i64_gt_s,
            0x56: self.i64_gt_u,
            0x57: self.i64_le_s,
            0x58: self.i64_le_u,
            0x59: self.i64_ge_s,
            0x5A: self.i64_ge_u,
            0x5B: self.f32_eq,
            0x5C: self.f32_ne,
            0x5D: self.f32_lt,
            0x5E: self.f32_gt,
            0x5F: self.f32_le,
            0x60: self.f32_ge,
            0x61: self.f64_eq,
            0x62: self.f64_ne,
            0x63: self.f64_lt,
            0x64: self.f64_gt,
            0x65: self.f64_le,
            0x66: self.f64_ge,
            0x67: self.i32_clz,
            0x68: self.i32_ctz,
            0x69: self.i32_popcnt,
            0x6A: self.i32_add,
            0x6B: self.i32_sub,
            0x6C: self.i32_mul,
            0x6D: self.i32_div_s,
            0x6E: self.i32_div_u,
            0x6F: self.i32_rem_s,
            0x70: self.i32_rem_u,
            0x71: self.i32_and,
            0x72: self.i32_or,
            0x73: self.i32_xor,
            0x74: self.i32_shl,
            0x75: self.i32_shr_s,
            0x76: self.i32_shr_u,
            0x77: self.i32_rotl,
            0x78: self.i32_rotr,
            0x79: self.i64_clz,
            0x7A: self.i64_ctz,
            0x7B: self.i64_popcnt,
            0x7C: self.i64_add,
            0x7D: self.i64_sub,
            0x7E: self.i64_mul,
            0x7F: self.i64_div_s,
            0x80: self.i64_div_u,
            0x81: self.i64_rem_s,
            0x82: self.i64_rem_u,
            0x83: self.i64_and,
            0x84: self.i64_or,
            0x85: self.i64_xor,
            0x86: self.i64_shl,
            0x87: self.i64_shr_s,
            0x88: self.i64_shr_u,
            0x89: self.i64_rotl,
            0x8A: self.i64_rotr,
            0x8B: self.f32_abs,
            0x8C: self.f32_neg,
            0x8D: self.f32_ceil,
            0x8E: self.f32_floor,
            0x8F: self.f32_trunc,
            0x90: self.f32_nearest,
            0x91: self.f32_sqrt,
            0x92: self.f32_add,
            0x93: self.f32_sub,
            0x94: self.f32_mul,
            0x95: self.f32_div,
            0x96: self.f32_min,
            0x97: self.f32_max,
            0x98: self.f32_copysign,
            0x99: self.f64_abs,
            0x9A: self.f64_neg,
            0x9B: self.f64_ceil,
            0x9C: self.f64_floor,
            0x9D: self.f64_trunc,
            0x9E: self.f64_nearest,
            0x9F: self.f64_sqrt,
            0xA0: self.f64_add,
            0xA1: self.f64_sub,
            0xA2: self.f64_mul,
            0xA3: self.f64_div,
            0xA4: self.f64_min,
            0xA5: self.f64_max,
            0xA6: self.f64_copysign,
            0xA7: self.i32_wrap_i64,
            0xA8: self.i32_trunc_s_f32,
            0xA9: self.i32_trunc_u_f32,
            0xAA: self.i32_trunc_s_f64,
            0xAB: self.i32_trunc_u_f64,
            0xAC: self.i64_extend_s_i32,
            0xAD: self.i64_extend_u_i32,
            0xAE: self.i64_trunc_s_f32,
            0xAF: self.i64_trunc_u_f32,
            0xB0: self.i64_trunc_s_f64,
            0xB1: self.i64_trunc_u_f64,
            0xB2: self.f32_convert_s_i32,
            0xB3: self.f32_convert_u_i32,
            0xB4: self.f32_convert_s_i64,
            0xB5: self.f32_convert_u_i64,
            0xB6: self.f32_demote_f64,
            0xB7: self.f64_convert_s_i32,
            0xB8: self.f64_convert_u_i32,
            0xB9: self.f64_convert_s_i64,
            0xBA: self.f64_convert_u_i64,
            0xBB: self.f64_promote_f32,
            0xBC: self.i32_reinterpret_f32,
            0xBD: self.i64_reinterpret_f64,
            0xBE: self.f32_reinterpret_i32,
            0xBF: self.f64_reinterpret_i64,
        }

        #: Constraint set to use for checking overflows and boundary conditions
        self.constraints = constraints

        self.zero_div = False
        self.overflow = False

        super().__init__()

    def __getstate__(self):
        state = super().__getstate__()
        state["mapping"] = self._mapping
        state["constraints"] = self.constraints
        state["zero_div"] = self.zero_div
        state["overflow"] = self.overflow
        return state

    def __setstate__(self, state):
        self._mapping = state["mapping"]
        self.constraints = state["constraints"]
        self.zero_div = state["zero_div"]
        self.overflow = state["overflow"]
        super().__setstate__(state)

    def check_overflow(self, expression) -> bool:
        if issymbolic(expression):
            self.overflow = Operators.OR(self.overflow, expression)
            return False
        return expression

    def check_zero_div(self, expression) -> bool:
        if issymbolic(expression):
            self.zero_div = Operators.OR(self.zero_div, expression)
            return False
        return expression

    def dispatch(self, inst, store, stack):
        """
        Selects the correct semantics for the given instruction, and executes them

        :param inst: the Instruction to execute
        :param store: the current Store
        :param stack: the current Stack
        :return: the result of the semantic function, which is (probably) always None
        """
        opcode = inst.opcode
        assert opcode in self._mapping
        func = self._mapping[opcode]
        try:
            if inst.imm:
                return func(store, stack, inst.imm)
            else:
                return func(store, stack)
        except (ZeroDivisionError, InvalidOperation):
            raise ZeroDivisionTrap()

    def unreachable(self, store, stack):
        raise UnreachableInstructionTrap()

    def nop(self, store, stack):
        pass

    def drop(self, store, stack):
        stack.has_type_on_top(Value_t, 1)
        stack.pop()

    def select(self, store, stack):
        c = stack.pop()
        v2 = stack.pop()
        v1 = stack.pop()
        assert isinstance(c, (I32, BitVec)), f"{type(c)} is not I32"
        if not issymbolic(v2) and not issymbolic(v1):
            assert type(v2) == type(v1), f"{type(v2)} is not the same as {type(v1)}"

        if issymbolic(c):
            stack.push(Operators.ITEBV(getattr(v1, "size", 32), c != 0, v1, v2))
        else:
            if c != 0:
                stack.push(v1)
            else:
                stack.push(v2)

    def get_local(self, store, stack, imm: LocalVarXsImm):
        f = stack.get_frame().frame
        assert imm.local_index in range(len(f.locals))
        self._publish("will_get_local", imm.local_index)
        stack.push(f.locals[imm.local_index])
        self._publish("did_get_local", imm.local_index, stack.peek())

    def set_local(self, store, stack, imm: LocalVarXsImm):
        f = stack.get_frame().frame
        assert imm.local_index in range(len(f.locals))
        stack.has_type_on_top(Value_t, 1)
        self._publish("will_set_local", imm.local_index, stack.peek())
        f.locals[imm.local_index] = stack.pop()
        self._publish("did_set_local", imm.local_index, f.locals[imm.local_index])

    def tee_local(self, store, stack, imm: LocalVarXsImm):
        stack.has_type_on_top(Value_t, 1)
        v = stack.pop()
        stack.push(v)
        stack.push(v)
        self.set_local(store, stack, imm)

    def get_global(self, store, stack, imm: GlobalVarXsImm):
        f = stack.get_frame().frame
        assert imm.global_index in range(len(f.module.globaladdrs))
        a = f.module.globaladdrs[imm.global_index]
        assert a in range(len(store.globals))
        glob = store.globals[a]
        self._publish("will_get_global", imm.global_index, glob.value)
        stack.push(glob.value)
        self._publish("did_get_global", imm.global_index, stack.peek())

    def set_global(self, store, stack, imm: GlobalVarXsImm):
        f = stack.get_frame().frame
        assert imm.global_index in range(len(f.module.globaladdrs))
        a = f.module.globaladdrs[imm.global_index]
        assert a in range(len(store.globals))
        stack.has_type_on_top(Value_t, 1)
        self._publish("will_set_global", imm.global_index, stack.peek())
        store.globals[a].value = stack.pop()
        self._publish("did_set_global", imm.global_index, store.globals[a].value)

    def i32_load(self, store, stack, imm: MemoryImm):
        f = stack.get_frame().frame
        assert f.module.memaddrs
        a = f.module.memaddrs[0]
        assert a in range(len(store.mems))
        mem = store.mems[a]
        stack.has_type_on_top(I32, 1)
        i = stack.pop()
        if issymbolic(i):
            raise ConcretizeStack(
                -1, I32, "Concretizing memory read", i
            )  # TODO - Implement a symbolic memory model
        ea = i + imm.offset
        if (ea + 4) - 1 not in mem:
            raise OutOfBoundsMemoryTrap(ea + 4)
        c = mem.read_int(ea, 32)
        stack.push(I32.cast(c))

    def i64_load(self, store, stack, imm: MemoryImm):
        f = stack.get_frame().frame
        assert f.module.memaddrs
        a = f.module.memaddrs[0]
        assert a in range(len(store.mems))
        mem = store.mems[a]
        stack.has_type_on_top(I32, 1)
        i = stack.pop()
        if issymbolic(i):
            raise ConcretizeStack(
                -1, I32, "Concretizing memory read", i
            )  # TODO - Implement a symbolic memory model
        ea = i + imm.offset
        if (ea + 8) - 1 not in mem:
            raise OutOfBoundsMemoryTrap(ea + 8)
        c = mem.read_int(ea, 64)
        stack.push(I64.cast(c))

    def int_load(self, store, stack, imm: MemoryImm, ty: type, size: int, signed: bool):
        assert ty in {I32, I64}, f"{type(ty)} is not an I32 or I64"
        f = stack.get_frame().frame
        assert f.module.memaddrs
        a = f.module.memaddrs[0]
        assert a in range(len(store.mems))
        mem = store.mems[a]
        stack.has_type_on_top(I32, 1)
        i = stack.pop()
        if issymbolic(i):
            raise ConcretizeStack(
                -1, I32, "Concretizing memory read", i
            )  # TODO - Implement a symbolic memory model
        ea = i + imm.offset
        if ea not in mem:
            raise OutOfBoundsMemoryTrap(ea)
        if ea + (size // 8) - 1 not in mem:
            raise OutOfBoundsMemoryTrap(ea + (size // 8))
        c = mem.read_int(ea, size)
        width = 32 if ty is I32 else 64
        if signed:
            c = Operators.SEXTEND(c, size, width)
        else:
            c = Operators.ZEXTEND(c, width)
        # Mypy can't figure out that that ty will definitely have a cast method, so we ignore the type
        stack.push(ty.cast(c))  # type: ignore

    def i32_load8_s(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I32, 8, True)

    def i32_load8_u(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I32, 8, False)

    def i32_load16_s(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I32, 16, True)

    def i32_load16_u(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I32, 16, False)

    def i64_load8_s(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I64, 8, True)

    def i64_load8_u(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I64, 8, False)

    def i64_load16_s(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I64, 16, True)

    def i64_load16_u(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I64, 16, False)

    def i64_load32_s(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I64, 32, True)

    def i64_load32_u(self, store, stack, imm: MemoryImm):
        self.int_load(store, stack, imm, I64, 32, False)

    def int_store(self, store, stack, imm: MemoryImm, ty: type, n=None):
        assert ty in {I32, I64}, f"{type(ty)} is not an I32 or I64"
        f = stack.get_frame().frame
        assert f.module.memaddrs
        a = f.module.memaddrs[0]
        assert a in range(len(store.mems))
        mem = store.mems[a]
        stack.has_type_on_top(ty, 1)
        c = stack.pop()
        stack.has_type_on_top(I32, 1)
        i = stack.pop()
        if issymbolic(i):
            raise ConcretizeStack(
                -2, I32, "Concretizing integer memory write", i
            )  # TODO - Implement a symbolic memory model
        ea = i + imm.offset
        N = n if n else (32 if ty is I32 else 64)
        mask = (1 << N) - 1
        if ea not in mem:
            raise OutOfBoundsMemoryTrap(ea)
        if (ea + (N // 8)) - 1 not in mem:
            raise OutOfBoundsMemoryTrap(ea + (N // 8))
        if n:
            b = [Operators.CHR(Operators.EXTRACT(c & mask, offset, 8)) for offset in range(0, N, 8)]
        else:
            b = [Operators.CHR(Operators.EXTRACT(c, offset, 8)) for offset in range(0, N, 8)]

        mem.write_bytes(ea, b)

    def i32_store(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I32)

    def i64_store(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I64)

    def i32_store8(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I32, 8)

    def i32_store16(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I32, 16)

    def i64_store8(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I64, 8)

    def i64_store16(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I64, 16)

    def i64_store32(self, store, stack, imm: MemoryImm):
        self.int_store(store, stack, imm, I64, 32)

    def current_memory(self, store, stack, imm: CurGrowMemImm):
        f = stack.get_frame().frame
        assert f.module.memaddrs
        a = f.module.memaddrs[0]
        assert a in range(len(store.mems))
        mem = store.mems[a]
        stack.push(I32(mem.npages))

    def grow_memory(self, store, stack, imm: CurGrowMemImm):
        f = stack.get_frame().frame
        assert f.module.memaddrs
        a = f.module.memaddrs[0]
        assert a in range(len(store.mems))
        mem = store.mems[a]
        sz = mem.npages
        stack.has_type_on_top(I32, 1)
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-1, I32, "Concretizing memory grow operand", stack.peek())
        if mem.grow(stack.pop()):
            stack.push(I32(sz))
        else:
            stack.push(I32(-1))

    def i32_const(self, store, stack, imm: I32ConstImm):
        stack.push(I32.cast(imm.value))

    def i64_const(self, store, stack, imm: I64ConstImm):
        stack.push(I64.cast(imm.value))

    def i32_eqz(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1 = stack.pop()
        v = c1 == 0
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_eq(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c2 == c1
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_ne(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c2 != c1
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_lt_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 < c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_lt_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.ULT(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_gt_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 > c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_gt_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.UGT(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_le_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 <= c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_le_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.ULE(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_ge_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 >= c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_ge_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.UGE(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_eqz(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1 = stack.pop()
        v = c1 == 0
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_eq(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c2 == c1
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_ne(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c2 != c1
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_lt_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 < c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_lt_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.ULT(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_gt_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 > c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_gt_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.UGT(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_le_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 <= c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_le_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.ULE(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_ge_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = c1 >= c2
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i64_ge_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        v = Operators.UGE(c1, c2)
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(I32.cast(I32(1) if v else I32(0)))

    def i32_clz(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1 = stack.pop()
        flag = Operators.EXTRACT(c1, 31, 1) == 1
        res = 0
        for pos in range(1, 32):
            res = Operators.ITEBV(32, flag, res, pos)
            flag = Operators.OR(flag, Operators.EXTRACT(c1, 31 - pos, 1) == 1)
        res = Operators.ITEBV(32, flag, res, 32)
        stack.push(I32.cast(res))

        # value = src.read()
        # flag = Operators.EXTRACT(value, 0, 1) == 1
        # res = 0
        # for pos in range(1, src.size):
        #     res = Operators.ITEBV(dest.size, flag, res, pos)
        #     flag = Operators.OR(flag, Operators.EXTRACT(value, pos, 1) == 1)
        #
        # cpu.CF = res == src.size
        # cpu.ZF = res == 0
        # dest.write(res)

    def i32_ctz(self, store, stack):  # Copied from x86 TZCNT
        stack.has_type_on_top(I32, 1)
        c1 = stack.pop()
        flag = Operators.EXTRACT(c1, 0, 1) == 1
        res = 0
        for pos in range(1, 32):
            res = Operators.ITEBV(32, flag, res, pos)
            flag = Operators.OR(flag, Operators.EXTRACT(c1, pos, 1) == 1)
        res = Operators.ITEBV(32, flag, res, 32)
        stack.push(I32.cast(res))

    def i32_popcnt(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1 = stack.pop()
        flag = Operators.EXTRACT(c1, 0, 1) != 0
        res = 0
        for pos in range(1, 32):
            res = Operators.ITEBV(32, flag, res + 1, res)
            flag = Operators.EXTRACT(c1, pos, 1) != 0
        res = Operators.ITEBV(32, flag, res + 1, res)
        stack.push(I32.cast(res))

    def i32_add(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast((c2 + c1) & MASK_32))

    def i32_sub(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast((c1 - c2 + 2 ** 32) & MASK_32))

    def i32_mul(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast((c2 * c1) & MASK_32))

    def i32_div_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        can_div_0 = c2 == 0
        if self.check_zero_div(can_div_0):
            raise ZeroDivisionTrap()
        res = Operators.SDIV(c1, c2)
        can_overflow = res == 2 ** 31
        if self.check_overflow(can_overflow):
            raise OverflowDivisionTrap()
        stack.push(I32.cast(res))

    def i32_div_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        can_div_0 = c2 == 0
        if self.check_zero_div(can_div_0):
            raise ZeroDivisionTrap()
        if not issymbolic(c2):
            c2 = I32.to_unsigned(c2)
        if not issymbolic(c1):
            c1 = I32.to_unsigned(c1)
        stack.push(I32.cast(Operators.UDIV(c1, c2)))

    def i32_rem_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if self.check_zero_div(c2 == 0):
            raise ZeroDivisionTrap()
        stack.push(I32.cast(Operators.SREM(c1, c2)))

    def i32_rem_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c2):
            c2 = I32.to_unsigned(c2)
        if not issymbolic(c1):
            c1 = I32.to_unsigned(c1)
        if self.check_zero_div(c2 == 0):
            raise ZeroDivisionTrap()
        stack.push(I32.cast(Operators.UREM(c1, c2)))

    def i32_and(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast(c2 & c1))

    def i32_or(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast(c2 | c1))

    def i32_xor(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast(c2 ^ c1))

    def i32_shl(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I32.cast((c1 << (c2 % 32)) & MASK_32))

    def i32_shr_s(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        k = c2 % 32
        stack.push(I32.cast(Operators.SAR(32, c1, k)))

    def i32_shr_u(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c2):
            c2 = I32.to_unsigned(c2)
        if not issymbolic(c1):
            c1 = I32.to_unsigned(c1)
        stack.push(I32.cast(c1 >> (c2 % 32)))

    def i32_rotl(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c1):
            c1 = I32.to_unsigned(c1)
        k = c2 % 32
        stack.push(I32.cast((c1 << k) | c1 >> (32 - k)))

    def i32_rotr(self, store, stack):
        stack.has_type_on_top(I32, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c1):
            c1 = I32.to_unsigned(c1)
        k = c2 % 32
        stack.push(I32.cast((c1 >> k) | c1 << (32 - k)))

    def i64_clz(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1 = stack.pop()
        flag = Operators.EXTRACT(c1, 63, 1) == 1
        res = 0
        for pos in range(1, 64):
            res = Operators.ITEBV(64, flag, res, pos)
            flag = Operators.OR(flag, Operators.EXTRACT(c1, 63 - pos, 1) == 1)

        res = Operators.ITEBV(64, flag, res, 64)
        stack.push(I64.cast(res))

    def i64_ctz(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1 = stack.pop()
        flag = Operators.EXTRACT(c1, 0, 1) == 1
        res = 0
        for pos in range(1, 64):
            res = Operators.ITEBV(64, flag, res, pos)
            flag = Operators.OR(flag, Operators.EXTRACT(c1, pos, 1) == 1)
        res = Operators.ITEBV(64, flag, res, 64)
        stack.push(I64.cast(res))

    def i64_popcnt(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1 = stack.pop()
        flag = Operators.EXTRACT(c1, 0, 1) != 0
        res = 0
        for pos in range(1, 64):
            res = Operators.ITEBV(64, flag, res + 1, res)
            flag = Operators.EXTRACT(c1, pos, 1) != 0
        res = Operators.ITEBV(64, flag, res + 1, res)
        stack.push(I64.cast(res))

    def i64_add(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast((c2 + c1) & MASK_64))

    def i64_sub(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast((c1 - c2 + 2 ** 64) & MASK_64))

    def i64_mul(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast((c2 * c1) & MASK_64))

    def i64_div_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        can_div_0 = c2 == 0
        if self.check_zero_div(can_div_0):
            raise ZeroDivisionTrap()
        if issymbolic(c1) or issymbolic(c2):
            res = Operators.SDIV(c1, c2)
        else:
            res = int(math.trunc(Decimal(c1) / Decimal(c2)))
        can_overflow = res == 2 ** 63
        if self.check_overflow(can_overflow):
            raise OverflowDivisionTrap()
        stack.push(I64.cast(res))

    def i64_div_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        can_div_0 = c2 == 0
        if self.check_zero_div(can_div_0):
            raise ZeroDivisionTrap()
        if not issymbolic(c2):
            c2 = I64.to_unsigned(c2)
        if not issymbolic(c1):
            c1 = I64.to_unsigned(c1)
        stack.push(I64.cast(Operators.UDIV(c1, c2)))

    def i64_rem_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if self.check_zero_div(c2 == 0):
            raise ZeroDivisionTrap()
        if issymbolic(c1) or issymbolic(c2):
            res = Operators.SREM(c1, c2)
        else:
            res = c1 - int(Decimal(c1) / Decimal(c2)) * c2
        stack.push(I64.cast(res))

    def i64_rem_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c2):
            c2 = I64.to_unsigned(c2)
        if not issymbolic(c1):
            c1 = I64.to_unsigned(c1)
        if self.check_zero_div(c2 == 0):
            raise ZeroDivisionTrap()
        stack.push(I64.cast(Operators.UREM(c1, c2)))

    def i64_and(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast(c2 & c1))

    def i64_or(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast(c2 | c1))

    def i64_xor(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast(c2 ^ c1))

    def i64_shl(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        stack.push(I64.cast((c1 << (c2 % 64)) & MASK_64))

    def i64_shr_s(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        k = c2 % 64
        stack.push(I64.cast(Operators.SAR(64, c1, k)))

    def i64_shr_u(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c2):
            c2 = I64.to_unsigned(c2)
        if not issymbolic(c1):
            c1 = I64.to_unsigned(c1)
        stack.push(I64.cast(c1 >> (c2 % 64)))

    def i64_rotl(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c1):
            c1 = I64.to_unsigned(c1)
        k = c2 % 64
        stack.push(I64.cast((c1 << k) | c1 >> (64 - k)))

    def i64_rotr(self, store, stack):
        stack.has_type_on_top(I64, 2)
        c2 = stack.pop()
        c1 = stack.pop()
        if not issymbolic(c1):
            c1 = I64.to_unsigned(c1)
        k = c2 % 64
        stack.push(I64.cast((c1 >> k) | c1 << (64 - k)))

    def i32_wrap_i64(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1: I64 = stack.pop()
        c1 &= MASK_32
        c1 = Operators.EXTRACT(c1, 0, 32)
        stack.push(I32.cast(c1))

    def i32_trunc_s_f32(self, store, stack):
        stack.has_type_on_top(F32, 1)
        c1: F32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F32, "Concretizing for float->int conversion", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I32, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I32, "infinity")
        if c1 >= 2 ** 31 or c1 <= -(2 ** 31) - 1:
            raise InvalidConversionTrap(I32, c1)
        stack.push(I32.cast(c1))

    def i32_trunc_u_f32(self, store, stack):
        stack.has_type_on_top(F32, 1)
        c1: F32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F32, "Concretizing for float->int conversion", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I32, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I32, "infinity")
        if c1 >= 2 ** 32 or c1 <= -1:
            raise InvalidConversionTrap(I32, c1)
        stack.push(I32.cast(c1))

    def i32_trunc_s_f64(self, store, stack):
        stack.has_type_on_top(F64, 1)
        c1: F64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F64, "Concretizing for float->int conversion", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I32, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I32, "infinity")
        if c1 >= 2 ** 31 or c1 <= -(2 ** 31) - 1:
            raise InvalidConversionTrap(I32, c1)
        stack.push(I32.cast(c1))

    def i32_trunc_u_f64(self, store, stack):
        stack.has_type_on_top(F64, 1)
        c1: F64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F64, "Concretizing for float->int conversion", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I32, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I32, "infinity")
        if c1 >= 2 ** 32 or c1 <= -1:
            raise InvalidConversionTrap(I32, c1)
        stack.push(I32.cast(c1))

    def i64_extend_s_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        stack.push(I64.cast(Operators.SEXTEND(c1, 32, 64)))

    def i64_extend_u_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        if issymbolic(
            c1
        ):  # ZEXTEND doesn't have a concept of sized ints, so it will promote a negative I32
            # to a negative I64 with the same value.
            stack.push(I64.cast(Operators.ZEXTEND(c1, 64)))
        else:
            stack.push(I64.cast(struct.unpack("q", bytes(c_int32(c1)) + b"\x00" * 4)[0]))

    def i64_trunc_s_f32(self, store, stack):
        stack.has_type_on_top(F32, 1)
        c1: F32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F32, "Concretizing float", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I64, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I64, "infinity")
        if c1 >= 2 ** 63 or c1 <= -(2 ** 63) - 1:
            raise InvalidConversionTrap(I64, c1)
        stack.push(I64.cast(c1))

    def i64_trunc_u_f32(self, store, stack):
        stack.has_type_on_top(F32, 1)
        c1: F32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F32, "Concretizing float", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I64, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I64, "infinity")
        if c1 >= 2 ** 64 or c1 <= -1:
            raise InvalidConversionTrap(I64, c1)
        stack.push(I64.cast(c1))

    def i64_trunc_s_f64(self, store, stack):
        stack.has_type_on_top(F64, 1)
        c1: F64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F64, "Concretizing float", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I64, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I64, "infinity")
        if c1 >= 2 ** 63 or c1 <= -(2 ** 63) - 1:
            raise InvalidConversionTrap(I64, c1)
        stack.push(I64.cast(c1))

    def i64_trunc_u_f64(self, store, stack):
        stack.has_type_on_top(F64, 1)
        c1: F64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F64, "Concretizing float", c1)
        if math.isnan(c1):
            raise InvalidConversionTrap(I64, "NaN")
        if math.isinf(c1):
            raise InvalidConversionTrap(I64, "infinity")
        if c1 >= 2 ** 64 or c1 <= -1:
            raise InvalidConversionTrap(I64, c1)
        stack.push(I64.cast(c1))

    def i32_reinterpret_f32(self, store, stack):
        stack.has_type_on_top(F32, 1)
        c1: F32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F32, "Concretizing float", c1)
        c1 = struct.unpack("i", struct.pack("f", c1))[0]
        stack.push(I32.cast(c1))

    def i64_reinterpret_f64(self, store, stack):
        stack.has_type_on_top(F64, 1)
        c1: F64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F64, "Concretizing float", c1)
        c1 = struct.unpack("q", struct.pack("d", c1))[0]
        stack.push(I64.cast(c1))

    ###########################################################################################################
    # Floating point instructions# Floating point instructions
    def float_load(self, store, stack, imm: MemoryImm, ty: type):
        assert ty in {F32, F64}, f"{type(ty)} is not an F32 or F64"
        size = 32 if ty == F32 else 64
        f = stack.get_frame().frame
        a = f.module.memaddrs[0]
        mem = store.mems[a]
        stack.has_type_on_top(I32, 1)
        i = stack.pop()
        if issymbolic(i):
            raise ConcretizeStack(
                -1, I32, "Concretizing float memory read", i
            )  # TODO - Implement a symbolic memory model
        ea = i + imm.offset
        if ea not in mem:
            raise OutOfBoundsMemoryTrap(ea)
        if (ea + (size // 8)) - 1 not in mem:
            raise OutOfBoundsMemoryTrap(ea + (size // 8))
        c = mem.read_int(ea, size)
        # Mypy can't figure out that that ty will definitely have a cast method, so we ignore the type
        ret = ty.cast(c)  # type: ignore
        stack.push(ret)

    def f32_load(self, store, stack, imm: MemoryImm):
        return self.float_load(store, stack, imm, F32)

    def f64_load(self, store, stack, imm: MemoryImm):
        return self.float_load(store, stack, imm, F64)

    def float_store(self, store, stack, imm: MemoryImm, ty: type, n=None):
        f = stack.get_frame().frame
        a = f.module.memaddrs[0]
        mem = store.mems[a]
        c = stack.pop()
        i = stack.pop()
        if issymbolic(i):
            raise ConcretizeStack(-2, I32, "Concretizing memory address for float_store", i)
        ea = i + imm.offset
        if ty == F32:
            size = 32
        else:
            size = 64
        if ea not in mem:
            raise OutOfBoundsMemoryTrap(ea)
        if (ea + (size // 8)) - 1 not in mem:
            raise OutOfBoundsMemoryTrap(ea + (size // 8))
        if not issymbolic(c):
            c = struct.unpack(
                "i" if size == 32 else "q", struct.pack("f" if size == 32 else "d", c)
            )[0]
        b = [Operators.CHR(Operators.EXTRACT(c, offset, 8)) for offset in range(0, size, 8)]
        mem.write_bytes(ea, b)

    def float_push_compare_return(self, stack, v, rettype=I32):
        if issymbolic(v):
            stack.push(Operators.ITEBV(32, v, I32(1), I32(0)))
        else:
            stack.push(rettype(v))

    def f32_store(self, store, stack, imm: MemoryImm):
        self.float_store(store, stack, imm, F32)

    def f64_store(self, store, stack, imm: MemoryImm):
        self.float_store(store, stack, imm, F64)

    def f32_const(self, store, stack, imm: F32ConstImm):
        stack.push(F32.cast(imm.value))

    def f64_const(self, store, stack, imm: F64ConstImm):
        stack.push(F64.cast(imm.value))

    def f32_unary(self, store, stack, op, rettype: type = I32):
        stack.has_type_on_top(F32, 1)
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-1, F32, "Concretizing before float op", stack.peek())
        v1 = stack.pop()
        v = op(v1)
        self.float_push_compare_return(stack, v, rettype)

    def f32_binary(self, store, stack, op, rettype: type = I32):
        stack.has_type_on_top(F32, 2)
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-1, F32, "Concretizing before float op", stack.peek())
        v2 = stack.pop()
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-2, F32, "Concretizing before float op", stack.peek())
        v1 = stack.pop()
        v = op(v1, v2)
        self.float_push_compare_return(stack, v, rettype)

    def f64_unary(self, store, stack, op, rettype: type = F64):
        stack.has_type_on_top(F64, 1)
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-1, F64, "Concretizing before float op", stack.peek())
        v1 = stack.pop()
        v = op(v1)
        self.float_push_compare_return(stack, v, rettype)

    def f64_binary(self, store, stack, op, rettype: type = I32):
        stack.has_type_on_top(F64, 2)
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-1, F64, "Concretizing before float op", stack.peek())
        v2 = stack.pop()
        if issymbolic(stack.peek()):
            raise ConcretizeStack(-2, F64, "Concretizing before float op", stack.peek())
        v1 = stack.pop()
        v = op(v1, v2)
        self.float_push_compare_return(stack, v, rettype)

    def f32_eq(self, store, stack):
        return self.f32_binary(store, stack, operator.eq)

    def f32_ne(self, store, stack):
        return self.f32_binary(store, stack, operator.ne)

    def f32_lt(self, store, stack):
        return self.f32_binary(store, stack, operator.lt)

    def f32_gt(self, store, stack):
        return self.f32_binary(store, stack, operator.gt)

    def f32_le(self, store, stack):
        return self.f32_binary(store, stack, operator.le)

    def f32_ge(self, store, stack):
        return self.f32_binary(store, stack, operator.ge)

    def f64_eq(self, store, stack):
        return self.f64_binary(store, stack, operator.eq)

    def f64_ne(self, store, stack):
        return self.f64_binary(store, stack, operator.ne)

    def f64_lt(self, store, stack):
        return self.f64_binary(store, stack, operator.lt)

    def f64_gt(self, store, stack):
        return self.f64_binary(store, stack, operator.gt)

    def f64_le(self, store, stack):
        return self.f64_binary(store, stack, operator.le)

    def f64_ge(self, store, stack):
        return self.f64_binary(store, stack, operator.ge)

    def f32_abs(self, store, stack):
        return self.f32_unary(store, stack, operator.abs, F32)

    def f32_neg(self, store, stack):
        return self.f32_unary(store, stack, operator.neg, F32)

    def f32_ceil(self, store, stack):
        return self.f32_unary(store, stack, operator_ceil, F32)

    def f32_floor(self, store, stack):
        return self.f32_unary(store, stack, operator_floor, F32)

    def f32_trunc(self, store, stack):
        return self.f32_unary(store, stack, operator_trunc, F32)

    def f32_nearest(self, store, stack):
        return self.f32_unary(store, stack, operator_nearest, F32)

    def f32_sqrt(self, store, stack):
        return self.f32_unary(store, stack, math.sqrt, F32)

    def f32_add(self, store, stack):
        return self.f32_binary(store, stack, operator.add, F32)

    def f32_sub(self, store, stack):
        return self.f32_binary(store, stack, operator.sub, F32)

    def f32_mul(self, store, stack):
        return self.f32_binary(store, stack, operator.mul, F32)

    def f32_div(self, store, stack):
        return self.f32_binary(store, stack, operator_div, F32)

    def f32_min(self, store, stack):
        return self.f32_binary(store, stack, operator_min, F32)

    def f32_max(self, store, stack):
        return self.f32_binary(store, stack, operator_max, F32)

    def f32_copysign(self, store, stack):
        return self.f32_binary(store, stack, math.copysign, F32)

    def f64_abs(self, store, stack):
        return self.f64_unary(store, stack, operator.abs, F64)

    def f64_neg(self, store, stack):
        return self.f64_unary(store, stack, operator.neg, F64)

    def f64_ceil(self, store, stack):
        return self.f64_unary(store, stack, operator_ceil, F64)

    def f64_floor(self, store, stack):
        return self.f64_unary(store, stack, operator_floor, F64)

    def f64_trunc(self, store, stack):
        return self.f64_unary(store, stack, operator_trunc, F64)

    def f64_nearest(self, store, stack):
        return self.f32_unary(store, stack, operator_nearest, F64)

    def f64_sqrt(self, store, stack):
        return self.f64_unary(store, stack, math.sqrt, F64)

    def f64_add(self, store, stack):
        return self.f64_binary(store, stack, operator.add, F64)

    def f64_sub(self, store, stack):
        return self.f64_binary(store, stack, operator.sub, F64)

    def f64_mul(self, store, stack):
        return self.f64_binary(store, stack, operator.mul, F64)

    def f64_div(self, store, stack):
        return self.f64_binary(store, stack, operator_div, F64)

    def f64_min(self, store, stack):
        return self.f64_binary(store, stack, operator_min, F64)

    def f64_max(self, store, stack):
        return self.f64_binary(store, stack, operator_max, F64)

    def f64_copysign(self, store, stack):
        return self.f64_binary(store, stack, math.copysign, F64)

    def f32_convert_s_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        stack.push(F32.cast(float(c1)))

    def f32_convert_u_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        stack.push(F32.cast(float(I32.to_unsigned(c1))))

    def f32_convert_s_i64(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1: I64 = stack.pop()
        stack.push(F32.cast(float(c1)))

    def f32_convert_u_i64(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1: I64 = stack.pop()
        stack.push(F32.cast(float(I64.to_unsigned(c1))))

    def f32_demote_f64(self, store, stack):
        stack.has_type_on_top(F64, 1)
        c1: F64 = stack.pop()
        if math.isnan(c1) or math.isinf(c1) or c1 == 0.0 or c1 == -0.0:
            stack.push(F32.cast(c1))
            return
        raise NotImplementedError("f32_demote_f64")
        # c1 = struct.unpack("f", struct.pack("d", c1)[:4])[0]
        # stack.push(F32.cast(c1))

    def f64_convert_s_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, I32, "Concretizing int for float conversion", c1)
        stack.push(F64.cast(float(c1)))

    def f64_convert_u_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, I32, "Concretizing int for float conversion", c1)
        stack.push(F64.cast(float(I32.to_unsigned(c1))))

    def f64_convert_s_i64(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1: I64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, I64, "Concretizing int for float conversion", c1)
        stack.push(F64.cast(float(c1)))

    def f64_convert_u_i64(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1: I64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, I64, "Concretizing int for float conversion", c1)
        stack.push(F64.cast(float(I64.to_unsigned(c1))))

    def f64_promote_f32(self, store, stack):
        stack.has_type_on_top(F32, 1)
        c1: F32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, F32, "Concretizing F32 for F64 promotion", c1)
        stack.push(F64.cast(c1))

    def f32_reinterpret_i32(self, store, stack):
        stack.has_type_on_top(I32, 1)
        c1: I32 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, I32, "Concretizing int for float conversion", c1)
        c1 = struct.unpack("f", struct.pack("i", c1))[0]
        stack.push(F32.cast(c1))

    def f64_reinterpret_i64(self, store, stack):
        stack.has_type_on_top(I64, 1)
        c1: I64 = stack.pop()
        if issymbolic(c1):
            raise ConcretizeStack(-1, I64, "Concretizing int for float conversion", c1)
        c1 = struct.unpack("d", struct.pack("q", c1))[0]
        stack.push(F64.cast(c1))


################################################################################################
# Unary and binary operators for floats that don't fit python
def operator_nearest(a):
    if math.isnan(a) or math.isinf(a):
        return a.integer
    else:
        return round(a)


def operator_trunc(a):
    if math.isnan(a) or math.isinf(a):
        return a.integer
    else:
        return math.trunc(a)


def operator_ceil(a):
    if math.isnan(a) or math.isinf(a):
        return a.integer
    else:
        return math.ceil(a)


def operator_floor(a):
    if math.isnan(a) or math.isinf(a):
        return a.integer
    else:
        return math.floor(a)


def operator_div(a, b):
    if b == 0:
        return math.inf
    else:
        return operator.truediv(a, b)


def operator_min(a, b):
    return a if a < b else b


def operator_max(a, b):
    return a if a > b else b



import itertools
import sys
import copy
from typing import Optional
from ...utils.helpers import PickleSerializer
from ...exceptions import SmtlibError
from .expression import (
    Expression,
    BitVecVariable,
    BoolVariable,
    ArrayVariable,
    Array,
    Bool,
    BitVec,
    BoolConstant,
    ArrayProxy,
    BoolEqual,
    Variable,
    Constant,
)
from .visitors import (
    GetDeclarations,
    TranslatorSmtlib,
    get_variables,
    simplify,
    replace,
    pretty_print,
)
from ...utils import config
import logging

logger = logging.getLogger(__name__)


class ConstraintException(SmtlibError):
    """
    Constraint exception
    """

    pass


class ConstraintSet:
    """ Constraint Sets

        An object containing a set of constraints. Serves also as a factory for
        new variables.
    """

    def __init__(self):
        self._constraints = list()
        self._parent = None
        self._sid = 0
        self._declarations = {}
        self._child = None

    def __reduce__(self):
        return (
            self.__class__,
            (),
            {
                "_parent": self._parent,
                "_constraints": self._constraints,
                "_sid": self._sid,
                "_declarations": self._declarations,
            },
        )

    def __hash__(self):
        return hash(self.constraints)

    def __enter__(self) -> "ConstraintSet":
        assert self._child is None
        self._child = self.__class__()
        self._child._parent = self
        self._child._sid = self._sid
        self._child._declarations = dict(self._declarations)
        return self._child

    def __exit__(self, ty, value, traceback) -> None:
        self._child._parent = None
        self._child = None

    def __len__(self) -> int:
        if self._parent is not None:
            return len(self._constraints) + len(self._parent)
        return len(self._constraints)

    def add(self, constraint) -> None:
        """
        Add a constraint to the set

        :param constraint: The constraint to add to the set.
        """
        if isinstance(constraint, bool):
            constraint = BoolConstant(constraint)
        assert isinstance(constraint, Bool)
        constraint = simplify(constraint)
        # If self._child is not None this constraint set has been forked and a
        # a derived constraintset may be using this. So we can't add any more
        # constraints to this one. After the child constraintSet is deleted
        # we regain the ability to add constraints.
        if self._child is not None:
            raise ConstraintException("ConstraintSet is frozen")

        if isinstance(constraint, BoolConstant):
            if not constraint.value:
                logger.info("Adding an impossible constant constraint")
                self._constraints = [constraint]
            else:
                return
        self._constraints.append(constraint)

    def _get_sid(self) -> int:
        """ Returns a unique id. """
        assert self._child is None
        self._sid += 1
        return self._sid

    def related_to(self, *related_to) -> "ConstraintSet":
        # sam.moelius: There is a flaw in how __get_related works: when called on certain
        # unsatisfiable sets, it can return a satisfiable one. The flaw arises when:
        #   * self consists of a single constraint C
        #   * C is the value of the related_to parameter
        #   * C contains no variables
        #   * C is unsatisfiable
        # Since C contains no variables, it is not considered "related to" itself and is thrown out
        # by __get_related. Since C was the sole element of self, __get_related returns the empty
        # set. Thus, __get_related was called on an unsatisfiable set, {C}, but it returned a
        # satisfiable one, {}.
        #   In light of the above, the core __get_related logic is currently disabled.
        """
        Slices this ConstraintSet keeping only the related constraints.
        Two constraints are independient if they can be expressed full using a
        disjoint set of variables.
        Todo: Research. constraints refering differen not overlapping parts of the same array
        should be considered independient.
        :param related_to: An expression
        :return:
        """

        if not related_to:
            return copy.copy(self)
        number_of_constraints = len(self.constraints)
        remaining_constraints = set(self.constraints)
        related_variables = set()
        for expression in related_to:
            related_variables |= get_variables(expression)
        related_constraints = set()

        added = True
        while added:
            added = False
            logger.debug("Related variables %r", [x.name for x in related_variables])
            for constraint in list(remaining_constraints):
                if isinstance(constraint, BoolConstant):
                    if constraint.value:
                        continue
                    else:
                        related_constraints = {constraint}
                        break

                variables = get_variables(constraint)
                if related_variables & variables or not (variables):
                    remaining_constraints.remove(constraint)
                    related_constraints.add(constraint)
                    related_variables |= variables
                    added = True

        logger.debug("Reduced %d constraints!!", number_of_constraints - len(related_constraints))
        # related_variables, related_constraints
        cs = ConstraintSet()
        for var in related_variables:
            cs._declare(var)
        for constraint in related_constraints:
            cs.add(constraint)
        return cs

    def to_string(self, replace_constants: bool = False) -> str:
        variables, constraints = self.get_declared_variables(), self.constraints

        if replace_constants:
            constant_bindings = {}
            for expression in constraints:
                if (
                    isinstance(expression, BoolEqual)
                    and isinstance(expression.operands[0], Variable)
                    and isinstance(expression.operands[1], (*Variable, *Constant))
                ):
                    constant_bindings[expression.operands[0]] = expression.operands[1]

        tmp = set()
        result = ""
        for var in variables:
            # FIXME
            # band aid hack around the fact that we are double declaring stuff :( :(
            if var.declaration in tmp:
                logger.warning("Variable '%s' was copied twice somewhere", var.name)
                continue
            tmp.add(var.declaration)
            result += var.declaration + "\n"

        translator = TranslatorSmtlib(use_bindings=True)
        for constraint in constraints:
            if replace_constants:
                constraint = simplify(replace(constraint, constant_bindings))
                # if no variables then it is a constant
                if isinstance(constraint, Constant) and constraint.value == True:
                    continue

            translator.visit(constraint)
        if replace_constants:
            for k, v in constant_bindings.items():
                translator.visit(k == v)

        for name, exp, smtlib in translator.bindings:
            if isinstance(exp, BitVec):
                result += f"(declare-fun {name} () (_ BitVec {exp.size}))"
            elif isinstance(exp, Bool):
                result += f"(declare-fun {name} () Bool)"
            elif isinstance(exp, Array):
                result += f"(declare-fun {name} () (Array (_ BitVec {exp.index_bits}) (_ BitVec {exp.value_bits})))"
            else:
                raise ConstraintException(f"Type not supported {exp!r}")
            result += f"(assert (= {name} {smtlib}))\n"

        constraint_str = translator.pop()
        while constraint_str is not None:
            if constraint_str != "true":
                result += f"(assert {constraint_str})\n"
            constraint_str = translator.pop()
        return result

    def _declare(self, var):
        """ Declare the variable `var` """
        if var.name in self._declarations:
            raise ValueError("Variable already declared")
        self._declarations[var.name] = var
        return var

    def get_declared_variables(self):
        """ Returns the variable expressions of this constraint set """
        return self._declarations.values()

    def get_variable(self, name):
        """ Returns the variable declared under name or None if it does not exists """
        return self._declarations.get(name)

    @property
    def declarations(self):
        """ Returns the variable expressions of this constraint set """
        declarations = GetDeclarations()
        for a in self.constraints:
            try:
                declarations.visit(a)
            except RuntimeError:
                # TODO: (defunct) move recursion management out of PickleSerializer
                if sys.getrecursionlimit() >= PickleSerializer.MAX_RECURSION:
                    raise ConstraintException(
                        f"declarations recursion limit surpassed {PickleSerializer.MAX_RECURSION}, aborting"
                    )
                new_limit = sys.getrecursionlimit() + PickleSerializer.DEFAULT_RECURSION
                if new_limit <= PickleSerializer.DEFAULT_RECURSION:
                    sys.setrecursionlimit(new_limit)
                    return self.declarations
        return declarations.result

    @property
    def constraints(self):
        """
        :rtype tuple
        :return: All constraints represented by this and parent sets.
        """
        if self._parent is not None:
            return tuple(self._constraints) + self._parent.constraints
        return tuple(self._constraints)

    def __iter__(self):
        return iter(self.constraints)

    def __str__(self):
        """ Returns a smtlib representation of the current state """
        return self.to_string()

    def _make_unique_name(self, name="VAR"):
        """ Makes a unique variable name"""
        # the while loop is necessary because appending the result of _get_sid()
        # is not guaranteed to make a unique name on the first try; a colliding
        # name could have been added previously
        while name in self._declarations:
            name = f"{name}_{self._get_sid()}"
        return name

    def is_declared(self, expression_var) -> bool:
        """ True if expression_var is declared in this constraint set """
        if not isinstance(expression_var, Variable):
            raise ValueError(f"Expression must be a Variable (not a {type(expression_var)})")
        return any(expression_var is x for x in self.get_declared_variables())

    def migrate(self, expression, name_migration_map=None):
        """ Migrate an expression created for a different constraint set to self.
            Returns an expression that can be used with this constraintSet

            All the foreign variables used in the expression are replaced by
            variables of this constraint set. If the variable was replaced before
            the replacement is taken from the provided migration map.

            The migration mapping is updated with new replacements.

            :param expression: the potentially foreign expression
            :param name_migration_map: mapping of already migrated variables. maps from string name of foreign variable to its currently existing migrated string name. this is updated during this migration.
            :return: a migrated expression where all the variables are local. name_migration_map is updated

        """
        if name_migration_map is None:
            name_migration_map = {}

        #  name_migration_map -> object_migration_map
        #  Based on the name mapping in name_migration_map build an object to
        #  object mapping to be used in the replacing of variables
        #  inv: object_migration_map's keys should ALWAYS be external/foreign
        #  expressions, and its values should ALWAYS be internal/local expressions
        object_migration_map = {}

        # List of foreign vars used in expression
        foreign_vars = itertools.filterfalse(self.is_declared, get_variables(expression))
        for foreign_var in foreign_vars:
            # If a variable with the same name was previously migrated
            if foreign_var.name in name_migration_map:
                migrated_name = name_migration_map[foreign_var.name]
                native_var = self.get_variable(migrated_name)
                assert (
                    native_var is not None
                ), "name_migration_map contains a variable that does not exist in this ConstraintSet"
                object_migration_map[foreign_var] = native_var
            else:
                # foreign_var was not found in the local declared variables nor
                # any variable with the same name was previously migrated
                # let's make a new unique internal name for it
                migrated_name = foreign_var.name
                if migrated_name in self._declarations:
                    migrated_name = self._make_unique_name(f"{foreign_var.name}_migrated")
                # Create and declare a new variable of given type
                if isinstance(foreign_var, Bool):
                    new_var = self.new_bool(name=migrated_name)
                elif isinstance(foreign_var, BitVec):
                    new_var = self.new_bitvec(foreign_var.size, name=migrated_name)
                elif isinstance(foreign_var, Array):
                    # Note that we are discarding the ArrayProxy encapsulation
                    new_var = self.new_array(
                        index_max=foreign_var.index_max,
                        index_bits=foreign_var.index_bits,
                        value_bits=foreign_var.value_bits,
                        name=migrated_name,
                    ).array
                else:
                    raise NotImplementedError(
                        f"Unknown expression type {type(foreign_var)} encountered during expression migration"
                    )
                # Update the var to var mapping
                object_migration_map[foreign_var] = new_var
                # Update the name to name mapping
                name_migration_map[foreign_var.name] = new_var.name

        #  Actually replace each appearance of migrated variables by the new ones
        migrated_expression = replace(expression, object_migration_map)
        return migrated_expression

    def new_bool(self, name=None, taint=frozenset(), avoid_collisions=False):
        """ Declares a free symbolic boolean in the constraint store
            :param name: try to assign name to internal variable representation,
                         if not unique, a numeric nonce will be appended
            :param avoid_collisions: potentially avoid_collisions the variable to avoid name collisions if True
            :return: a fresh BoolVariable
        """
        if name is None:
            name = "B"
            avoid_collisions = True
        if avoid_collisions:
            name = self._make_unique_name(name)
        if not avoid_collisions and name in self._declarations:
            raise ValueError(f"Name {name} already used")
        var = BoolVariable(name, taint=taint)
        return self._declare(var)

    def new_bitvec(self, size, name=None, taint=frozenset(), avoid_collisions=False):
        """ Declares a free symbolic bitvector in the constraint store
            :param size: size in bits for the bitvector
            :param name: try to assign name to internal variable representation,
                         if not unique, a numeric nonce will be appended
            :param avoid_collisions: potentially avoid_collisions the variable to avoid name collisions if True
            :return: a fresh BitVecVariable
        """
        if size <= 0:
            raise ValueError(f"Bitvec size ({size}) can't be equal to or less than 0")
        if name is None:
            name = "BV"
            avoid_collisions = True
        if avoid_collisions:
            name = self._make_unique_name(name)
        if not avoid_collisions and name in self._declarations:
            raise ValueError(f"Name {name} already used")
        var = BitVecVariable(size, name, taint=taint)
        return self._declare(var)

    def new_array(
        self,
        index_bits=32,
        name=None,
        index_max=None,
        value_bits=8,
        taint=frozenset(),
        avoid_collisions=False,
        default=None,
    ):
        """ Declares a free symbolic array of value_bits long bitvectors in the constraint store.
            :param index_bits: size in bits for the array indexes one of [32, 64]
            :param value_bits: size in bits for the array values
            :param name: try to assign name to internal variable representation,
                         if not unique, a numeric nonce will be appended
            :param index_max: upper limit for indexes on this array (#FIXME)
            :param avoid_collisions: potentially avoid_collisions the variable to avoid name collisions if True
            :param default: default for not initialized values
            :return: a fresh ArrayProxy
        """
        if name is None:
            name = "A"
            avoid_collisions = True
        if avoid_collisions:
            name = self._make_unique_name(name)
        if not avoid_collisions and name in self._declarations:
            raise ValueError(f"Name {name} already used")
        var = self._declare(ArrayVariable(index_bits, index_max, value_bits, name, taint=taint))
        return ArrayProxy(var, default=default)



#
#
# AUTOGENERATED, DO NOT EDIT
#
# From version: 4.11
#

i386 = {
    0: "sys_restart_syscall",
    1: "sys_exit",
    2: "sys_fork",
    3: "sys_read",
    4: "sys_write",
    5: "sys_open",
    6: "sys_close",
    7: "sys_waitpid",
    8: "sys_creat",
    9: "sys_link",
    10: "sys_unlink",
    11: "sys_execve",
    12: "sys_chdir",
    13: "sys_time",
    14: "sys_mknod",
    15: "sys_chmod",
    16: "sys_lchown16",
    18: "sys_stat",
    19: "sys_lseek",
    20: "sys_getpid",
    21: "sys_mount",
    22: "sys_oldumount",
    23: "sys_setuid16",
    24: "sys_getuid16",
    25: "sys_stime",
    26: "sys_ptrace",
    27: "sys_alarm",
    28: "sys_fstat",
    29: "sys_pause",
    30: "sys_utime",
    33: "sys_access",
    34: "sys_nice",
    36: "sys_sync",
    37: "sys_kill",
    38: "sys_rename",
    39: "sys_mkdir",
    40: "sys_rmdir",
    41: "sys_dup",
    42: "sys_pipe",
    43: "sys_times",
    45: "sys_brk",
    46: "sys_setgid16",
    47: "sys_getgid16",
    48: "sys_signal",
    49: "sys_geteuid16",
    50: "sys_getegid16",
    51: "sys_acct",
    52: "sys_umount",
    54: "sys_ioctl",
    55: "sys_fcntl",
    57: "sys_setpgid",
    59: "sys_olduname",
    60: "sys_umask",
    61: "sys_chroot",
    62: "sys_ustat",
    63: "sys_dup2",
    64: "sys_getppid",
    65: "sys_getpgrp",
    66: "sys_setsid",
    67: "sys_sigaction",
    68: "sys_sgetmask",
    69: "sys_ssetmask",
    70: "sys_setreuid16",
    71: "sys_setregid16",
    72: "sys_sigsuspend",
    73: "sys_sigpending",
    74: "sys_sethostname",
    75: "sys_setrlimit",
    76: "sys_old_getrlimit",
    77: "sys_getrusage",
    78: "sys_gettimeofday",
    79: "sys_settimeofday",
    80: "sys_getgroups16",
    81: "sys_setgroups16",
    82: "sys_old_select",
    83: "sys_symlink",
    84: "sys_lstat",
    85: "sys_readlink",
    86: "sys_uselib",
    87: "sys_swapon",
    88: "sys_reboot",
    89: "sys_old_readdir",
    90: "sys_old_mmap",
    91: "sys_munmap",
    92: "sys_truncate",
    93: "sys_ftruncate",
    94: "sys_fchmod",
    95: "sys_fchown16",
    96: "sys_getpriority",
    97: "sys_setpriority",
    99: "sys_statfs",
    100: "sys_fstatfs",
    101: "sys_ioperm",
    102: "sys_socketcall",
    103: "sys_syslog",
    104: "sys_setitimer",
    105: "sys_getitimer",
    106: "sys_newstat",
    107: "sys_newlstat",
    108: "sys_newfstat",
    109: "sys_uname",
    110: "sys_iopl",
    111: "sys_vhangup",
    113: "sys_vm86old",
    114: "sys_wait4",
    115: "sys_swapoff",
    116: "sys_sysinfo",
    117: "sys_ipc",
    118: "sys_fsync",
    119: "sys_sigreturn",
    120: "sys_clone",
    121: "sys_setdomainname",
    122: "sys_newuname",
    123: "sys_modify_ldt",
    124: "sys_adjtimex",
    125: "sys_mprotect",
    126: "sys_sigprocmask",
    128: "sys_init_module",
    129: "sys_delete_module",
    131: "sys_quotactl",
    132: "sys_getpgid",
    133: "sys_fchdir",
    134: "sys_bdflush",
    135: "sys_sysfs",
    136: "sys_personality",
    138: "sys_setfsuid16",
    139: "sys_setfsgid16",
    140: "sys_llseek",
    141: "sys_getdents",
    142: "sys_select",
    143: "sys_flock",
    144: "sys_msync",
    145: "sys_readv",
    146: "sys_writev",
    147: "sys_getsid",
    148: "sys_fdatasync",
    149: "sys_sysctl",
    150: "sys_mlock",
    151: "sys_munlock",
    152: "sys_mlockall",
    153: "sys_munlockall",
    154: "sys_sched_setparam",
    155: "sys_sched_getparam",
    156: "sys_sched_setscheduler",
    157: "sys_sched_getscheduler",
    158: "sys_sched_yield",
    159: "sys_sched_get_priority_max",
    160: "sys_sched_get_priority_min",
    161: "sys_sched_rr_get_interval",
    162: "sys_nanosleep",
    163: "sys_mremap",
    164: "sys_setresuid16",
    165: "sys_getresuid16",
    166: "sys_vm86",
    168: "sys_poll",
    170: "sys_setresgid16",
    171: "sys_getresgid16",
    172: "sys_prctl",
    173: "sys_rt_sigreturn",
    174: "sys_rt_sigaction",
    175: "sys_rt_sigprocmask",
    176: "sys_rt_sigpending",
    177: "sys_rt_sigtimedwait",
    178: "sys_rt_sigqueueinfo",
    179: "sys_rt_sigsuspend",
    180: "sys_pread64",
    181: "sys_pwrite64",
    182: "sys_chown16",
    183: "sys_getcwd",
    184: "sys_capget",
    185: "sys_capset",
    186: "sys_sigaltstack",
    187: "sys_sendfile",
    190: "sys_vfork",
    191: "sys_getrlimit",
    192: "sys_mmap_pgoff",
    193: "sys_truncate64",
    194: "sys_ftruncate64",
    195: "sys_stat64",
    196: "sys_lstat64",
    197: "sys_fstat64",
    198: "sys_lchown",
    199: "sys_getuid",
    200: "sys_getgid",
    201: "sys_geteuid",
    202: "sys_getegid",
    203: "sys_setreuid",
    204: "sys_setregid",
    205: "sys_getgroups",
    206: "sys_setgroups",
    207: "sys_fchown",
    208: "sys_setresuid",
    209: "sys_getresuid",
    210: "sys_setresgid",
    211: "sys_getresgid",
    212: "sys_chown",
    213: "sys_setuid",
    214: "sys_setgid",
    215: "sys_setfsuid",
    216: "sys_setfsgid",
    217: "sys_pivot_root",
    218: "sys_mincore",
    219: "sys_madvise",
    220: "sys_getdents64",
    221: "sys_fcntl64",
    224: "sys_gettid",
    225: "sys_readahead",
    226: "sys_setxattr",
    227: "sys_lsetxattr",
    228: "sys_fsetxattr",
    229: "sys_getxattr",
    230: "sys_lgetxattr",
    231: "sys_fgetxattr",
    232: "sys_listxattr",
    233: "sys_llistxattr",
    234: "sys_flistxattr",
    235: "sys_removexattr",
    236: "sys_lremovexattr",
    237: "sys_fremovexattr",
    238: "sys_tkill",
    239: "sys_sendfile64",
    240: "sys_futex",
    241: "sys_sched_setaffinity",
    242: "sys_sched_getaffinity",
    243: "sys_set_thread_area",
    244: "sys_get_thread_area",
    245: "sys_io_setup",
    246: "sys_io_destroy",
    247: "sys_io_getevents",
    248: "sys_io_submit",
    249: "sys_io_cancel",
    250: "sys_fadvise64",
    252: "sys_exit_group",
    253: "sys_lookup_dcookie",
    254: "sys_epoll_create",
    255: "sys_epoll_ctl",
    256: "sys_epoll_wait",
    257: "sys_remap_file_pages",
    258: "sys_set_tid_address",
    259: "sys_timer_create",
    260: "sys_timer_settime",
    261: "sys_timer_gettime",
    262: "sys_timer_getoverrun",
    263: "sys_timer_delete",
    264: "sys_clock_settime",
    265: "sys_clock_gettime",
    266: "sys_clock_getres",
    267: "sys_clock_nanosleep",
    268: "sys_statfs64",
    269: "sys_fstatfs64",
    270: "sys_tgkill",
    271: "sys_utimes",
    272: "sys_fadvise64_64",
    274: "sys_mbind",
    275: "sys_get_mempolicy",
    276: "sys_set_mempolicy",
    277: "sys_mq_open",
    278: "sys_mq_unlink",
    279: "sys_mq_timedsend",
    280: "sys_mq_timedreceive",
    281: "sys_mq_notify",
    282: "sys_mq_getsetattr",
    283: "sys_kexec_load",
    284: "sys_waitid",
    286: "sys_add_key",
    287: "sys_request_key",
    288: "sys_keyctl",
    289: "sys_ioprio_set",
    290: "sys_ioprio_get",
    291: "sys_inotify_init",
    292: "sys_inotify_add_watch",
    293: "sys_inotify_rm_watch",
    294: "sys_migrate_pages",
    295: "sys_openat",
    296: "sys_mkdirat",
    297: "sys_mknodat",
    298: "sys_fchownat",
    299: "sys_futimesat",
    300: "sys_fstatat64",
    301: "sys_unlinkat",
    302: "sys_renameat",
    303: "sys_linkat",
    304: "sys_symlinkat",
    305: "sys_readlinkat",
    306: "sys_fchmodat",
    307: "sys_faccessat",
    308: "sys_pselect6",
    309: "sys_ppoll",
    310: "sys_unshare",
    311: "sys_set_robust_list",
    312: "sys_get_robust_list",
    313: "sys_splice",
    314: "sys_sync_file_range",
    315: "sys_tee",
    316: "sys_vmsplice",
    317: "sys_move_pages",
    318: "sys_getcpu",
    319: "sys_epoll_pwait",
    320: "sys_utimensat",
    321: "sys_signalfd",
    322: "sys_timerfd_create",
    323: "sys_eventfd",
    324: "sys_fallocate",
    325: "sys_timerfd_settime",
    326: "sys_timerfd_gettime",
    327: "sys_signalfd4",
    328: "sys_eventfd2",
    329: "sys_epoll_create1",
    330: "sys_dup3",
    331: "sys_pipe2",
    332: "sys_inotify_init1",
    333: "sys_preadv",
    334: "sys_pwritev",
    335: "sys_rt_tgsigqueueinfo",
    336: "sys_perf_event_open",
    337: "sys_recvmmsg",
    338: "sys_fanotify_init",
    339: "sys_fanotify_mark",
    340: "sys_prlimit64",
    341: "sys_name_to_handle_at",
    342: "sys_open_by_handle_at",
    343: "sys_clock_adjtime",
    344: "sys_syncfs",
    345: "sys_sendmmsg",
    346: "sys_setns",
    347: "sys_process_vm_readv",
    348: "sys_process_vm_writev",
    349: "sys_kcmp",
    350: "sys_finit_module",
    351: "sys_sched_setattr",
    352: "sys_sched_getattr",
    353: "sys_renameat2",
    354: "sys_seccomp",
    355: "sys_getrandom",
    356: "sys_memfd_create",
    357: "sys_bpf",
    358: "sys_execveat",
    359: "sys_socket",
    360: "sys_socketpair",
    361: "sys_bind",
    362: "sys_connect",
    363: "sys_listen",
    364: "sys_accept4",
    365: "sys_getsockopt",
    366: "sys_setsockopt",
    367: "sys_getsockname",
    368: "sys_getpeername",
    369: "sys_sendto",
    370: "sys_sendmsg",
    371: "sys_recvfrom",
    372: "sys_recvmsg",
    373: "sys_shutdown",
    374: "sys_userfaultfd",
    375: "sys_membarrier",
    376: "sys_mlock2",
    377: "sys_copy_file_range",
    378: "sys_preadv2",
    379: "sys_pwritev2",
    380: "sys_pkey_mprotect",
    381: "sys_pkey_alloc",
    382: "sys_pkey_free",
    383: "sys_statx",
}
amd64 = {
    0: "sys_read",
    1: "sys_write",
    2: "sys_open",
    3: "sys_close",
    4: "sys_newstat",
    5: "sys_newfstat",
    6: "sys_newlstat",
    7: "sys_poll",
    8: "sys_lseek",
    9: "sys_mmap",
    10: "sys_mprotect",
    11: "sys_munmap",
    12: "sys_brk",
    13: "sys_rt_sigaction",
    14: "sys_rt_sigprocmask",
    15: "sys_rt_sigreturn/ptregs",
    16: "sys_ioctl",
    17: "sys_pread64",
    18: "sys_pwrite64",
    19: "sys_readv",
    20: "sys_writev",
    21: "sys_access",
    22: "sys_pipe",
    23: "sys_select",
    24: "sys_sched_yield",
    25: "sys_mremap",
    26: "sys_msync",
    27: "sys_mincore",
    28: "sys_madvise",
    29: "sys_shmget",
    30: "sys_shmat",
    31: "sys_shmctl",
    32: "sys_dup",
    33: "sys_dup2",
    34: "sys_pause",
    35: "sys_nanosleep",
    36: "sys_getitimer",
    37: "sys_alarm",
    38: "sys_setitimer",
    39: "sys_getpid",
    40: "sys_sendfile64",
    41: "sys_socket",
    42: "sys_connect",
    43: "sys_accept",
    44: "sys_sendto",
    45: "sys_recvfrom",
    46: "sys_sendmsg",
    47: "sys_recvmsg",
    48: "sys_shutdown",
    49: "sys_bind",
    50: "sys_listen",
    51: "sys_getsockname",
    52: "sys_getpeername",
    53: "sys_socketpair",
    54: "sys_setsockopt",
    55: "sys_getsockopt",
    56: "sys_clone_ptregs",
    57: "sys_fork/ptregs",
    58: "sys_vfork/ptregs",
    59: "sys_execve/ptregs",
    60: "sys_exit",
    61: "sys_wait4",
    62: "sys_kill",
    63: "sys_newuname",
    64: "sys_semget",
    65: "sys_semop",
    66: "sys_semctl",
    67: "sys_shmdt",
    68: "sys_msgget",
    69: "sys_msgsnd",
    70: "sys_msgrcv",
    71: "sys_msgctl",
    72: "sys_fcntl",
    73: "sys_flock",
    74: "sys_fsync",
    75: "sys_fdatasync",
    76: "sys_truncate",
    77: "sys_ftruncate",
    78: "sys_getdents",
    79: "sys_getcwd",
    80: "sys_chdir",
    81: "sys_fchdir",
    82: "sys_rename",
    83: "sys_mkdir",
    84: "sys_rmdir",
    85: "sys_creat",
    86: "sys_link",
    87: "sys_unlink",
    88: "sys_symlink",
    89: "sys_readlink",
    90: "sys_chmod",
    91: "sys_fchmod",
    92: "sys_chown",
    93: "sys_fchown",
    94: "sys_lchown",
    95: "sys_umask",
    96: "sys_gettimeofday",
    97: "sys_getrlimit",
    98: "sys_getrusage",
    99: "sys_sysinfo",
    100: "sys_times",
    101: "sys_ptrace",
    102: "sys_getuid",
    103: "sys_syslog",
    104: "sys_getgid",
    105: "sys_setuid",
    106: "sys_setgid",
    107: "sys_geteuid",
    108: "sys_getegid",
    109: "sys_setpgid",
    110: "sys_getppid",
    111: "sys_getpgrp",
    112: "sys_setsid",
    113: "sys_setreuid",
    114: "sys_setregid",
    115: "sys_getgroups",
    116: "sys_setgroups",
    117: "sys_setresuid",
    118: "sys_getresuid",
    119: "sys_setresgid",
    120: "sys_getresgid",
    121: "sys_getpgid",
    122: "sys_setfsuid",
    123: "sys_setfsgid",
    124: "sys_getsid",
    125: "sys_capget",
    126: "sys_capset",
    127: "sys_rt_sigpending",
    128: "sys_rt_sigtimedwait",
    129: "sys_rt_sigqueueinfo",
    130: "sys_rt_sigsuspend",
    131: "sys_sigaltstack",
    132: "sys_utime",
    133: "sys_mknod",
    135: "sys_personality",
    136: "sys_ustat",
    137: "sys_statfs",
    138: "sys_fstatfs",
    139: "sys_sysfs",
    140: "sys_getpriority",
    141: "sys_setpriority",
    142: "sys_sched_setparam",
    143: "sys_sched_getparam",
    144: "sys_sched_setscheduler",
    145: "sys_sched_getscheduler",
    146: "sys_sched_get_priority_max",
    147: "sys_sched_get_priority_min",
    148: "sys_sched_rr_get_interval",
    149: "sys_mlock",
    150: "sys_munlock",
    151: "sys_mlockall",
    152: "sys_munlockall",
    153: "sys_vhangup",
    154: "sys_modify_ldt",
    155: "sys_pivot_root",
    156: "sys_sysctl",
    157: "sys_prctl",
    158: "sys_arch_prctl",
    159: "sys_adjtimex",
    160: "sys_setrlimit",
    161: "sys_chroot",
    162: "sys_sync",
    163: "sys_acct",
    164: "sys_settimeofday",
    165: "sys_mount",
    166: "sys_umount",
    167: "sys_swapon",
    168: "sys_swapoff",
    169: "sys_reboot",
    170: "sys_sethostname",
    171: "sys_setdomainname",
    172: "sys_iopl/ptregs",
    173: "sys_ioperm",
    175: "sys_init_module",
    176: "sys_delete_module",
    179: "sys_quotactl",
    186: "sys_gettid",
    187: "sys_readahead",
    188: "sys_setxattr",
    189: "sys_lsetxattr",
    190: "sys_fsetxattr",
    191: "sys_getxattr",
    192: "sys_lgetxattr",
    193: "sys_fgetxattr",
    194: "sys_listxattr",
    195: "sys_llistxattr",
    196: "sys_flistxattr",
    197: "sys_removexattr",
    198: "sys_lremovexattr",
    199: "sys_fremovexattr",
    200: "sys_tkill",
    201: "sys_time",
    202: "sys_futex",
    203: "sys_sched_setaffinity",
    204: "sys_sched_getaffinity",
    206: "sys_io_setup",
    207: "sys_io_destroy",
    208: "sys_io_getevents",
    209: "sys_io_submit",
    210: "sys_io_cancel",
    212: "sys_lookup_dcookie",
    213: "sys_epoll_create",
    216: "sys_remap_file_pages",
    217: "sys_getdents64",
    218: "sys_set_tid_address",
    219: "sys_restart_syscall",
    220: "sys_semtimedop",
    221: "sys_fadvise64",
    222: "sys_timer_create",
    223: "sys_timer_settime",
    224: "sys_timer_gettime",
    225: "sys_timer_getoverrun",
    226: "sys_timer_delete",
    227: "sys_clock_settime",
    228: "sys_clock_gettime",
    229: "sys_clock_getres",
    230: "sys_clock_nanosleep",
    231: "sys_exit_group",
    232: "sys_epoll_wait",
    233: "sys_epoll_ctl",
    234: "sys_tgkill",
    235: "sys_utimes",
    237: "sys_mbind",
    238: "sys_set_mempolicy",
    239: "sys_get_mempolicy",
    240: "sys_mq_open",
    241: "sys_mq_unlink",
    242: "sys_mq_timedsend",
    243: "sys_mq_timedreceive",
    244: "sys_mq_notify",
    245: "sys_mq_getsetattr",
    246: "sys_kexec_load",
    247: "sys_waitid",
    248: "sys_add_key",
    249: "sys_request_key",
    250: "sys_keyctl",
    251: "sys_ioprio_set",
    252: "sys_ioprio_get",
    253: "sys_inotify_init",
    254: "sys_inotify_add_watch",
    255: "sys_inotify_rm_watch",
    256: "sys_migrate_pages",
    257: "sys_openat",
    258: "sys_mkdirat",
    259: "sys_mknodat",
    260: "sys_fchownat",
    261: "sys_futimesat",
    262: "sys_newfstatat",
    263: "sys_unlinkat",
    264: "sys_renameat",
    265: "sys_linkat",
    266: "sys_symlinkat",
    267: "sys_readlinkat",
    268: "sys_fchmodat",
    269: "sys_faccessat",
    270: "sys_pselect6",
    271: "sys_ppoll",
    272: "sys_unshare",
    273: "sys_set_robust_list",
    274: "sys_get_robust_list",
    275: "sys_splice",
    276: "sys_tee",
    277: "sys_sync_file_range",
    278: "sys_vmsplice",
    279: "sys_move_pages",
    280: "sys_utimensat",
    281: "sys_epoll_pwait",
    282: "sys_signalfd",
    283: "sys_timerfd_create",
    284: "sys_eventfd",
    285: "sys_fallocate",
    286: "sys_timerfd_settime",
    287: "sys_timerfd_gettime",
    288: "sys_accept4",
    289: "sys_signalfd4",
    290: "sys_eventfd2",
    291: "sys_epoll_create1",
    292: "sys_dup3",
    293: "sys_pipe2",
    294: "sys_inotify_init1",
    295: "sys_preadv",
    296: "sys_pwritev",
    297: "sys_rt_tgsigqueueinfo",
    298: "sys_perf_event_open",
    299: "sys_recvmmsg",
    300: "sys_fanotify_init",
    301: "sys_fanotify_mark",
    302: "sys_prlimit64",
    303: "sys_name_to_handle_at",
    304: "sys_open_by_handle_at",
    305: "sys_clock_adjtime",
    306: "sys_syncfs",
    307: "sys_sendmmsg",
    308: "sys_setns",
    309: "sys_getcpu",
    310: "sys_process_vm_readv",
    311: "sys_process_vm_writev",
    312: "sys_kcmp",
    313: "sys_finit_module",
    314: "sys_sched_setattr",
    315: "sys_sched_getattr",
    316: "sys_renameat2",
    317: "sys_seccomp",
    318: "sys_getrandom",
    319: "sys_memfd_create",
    320: "sys_kexec_file_load",
    321: "sys_bpf",
    322: "sys_execveat/ptregs",
    323: "sys_userfaultfd",
    324: "sys_membarrier",
    325: "sys_mlock2",
    326: "sys_copy_file_range",
    327: "sys_preadv2",
    328: "sys_pwritev2",
    329: "sys_pkey_mprotect",
    330: "sys_pkey_alloc",
    331: "sys_pkey_free",
    332: "sys_statx",
    512: "compat_sys_rt_sigaction",
    513: "sys32_x32_rt_sigreturn",
    514: "compat_sys_ioctl",
    515: "compat_sys_readv",
    516: "compat_sys_writev",
    517: "compat_sys_recvfrom",
    518: "compat_sys_sendmsg",
    519: "compat_sys_recvmsg",
    520: "compat_sys_execve/ptregs",
    521: "compat_sys_ptrace",
    522: "compat_sys_rt_sigpending",
    523: "compat_sys_rt_sigtimedwait",
    524: "compat_sys_rt_sigqueueinfo",
    525: "compat_sys_sigaltstack",
    526: "compat_sys_timer_create",
    527: "compat_sys_mq_notify",
    528: "compat_sys_kexec_load",
    529: "compat_sys_waitid",
    530: "compat_sys_set_robust_list",
    531: "compat_sys_get_robust_list",
    532: "compat_sys_vmsplice",
    533: "compat_sys_move_pages",
    534: "compat_sys_preadv64",
    535: "compat_sys_pwritev64",
    536: "compat_sys_rt_tgsigqueueinfo",
    537: "compat_sys_recvmmsg",
    538: "compat_sys_sendmmsg",
    539: "compat_sys_process_vm_readv",
    540: "compat_sys_process_vm_writev",
    541: "compat_sys_setsockopt",
    542: "compat_sys_getsockopt",
    543: "compat_sys_io_setup",
    544: "compat_sys_io_submit",
    545: "compat_sys_execveat/ptregs",
    546: "compat_sys_preadv64v2",
    547: "compat_sys_pwritev64v2",
}
armv7 = {
    0: "sys_restart_syscall",
    1: "sys_exit",
    2: "sys_fork",
    3: "sys_read",
    4: "sys_write",
    5: "sys_open",
    6: "sys_close",
    8: "sys_creat",
    9: "sys_link",
    10: "sys_unlink",
    11: "sys_execve",
    12: "sys_chdir",
    13: "sys_time",
    14: "sys_mknod",
    15: "sys_chmod",
    16: "sys_lchown16",
    19: "sys_lseek",
    20: "sys_getpid",
    21: "sys_mount",
    22: "sys_oldumount",
    23: "sys_setuid16",
    24: "sys_getuid16",
    25: "sys_stime",
    26: "sys_ptrace",
    27: "sys_alarm",
    29: "sys_pause",
    30: "sys_utime",
    33: "sys_access",
    34: "sys_nice",
    36: "sys_sync",
    37: "sys_kill",
    38: "sys_rename",
    39: "sys_mkdir",
    40: "sys_rmdir",
    41: "sys_dup",
    42: "sys_pipe",
    43: "sys_times",
    45: "sys_brk",
    46: "sys_setgid16",
    47: "sys_getgid16",
    49: "sys_geteuid16",
    50: "sys_getegid16",
    51: "sys_acct",
    52: "sys_umount",
    54: "sys_ioctl",
    55: "sys_fcntl",
    57: "sys_setpgid",
    60: "sys_umask",
    61: "sys_chroot",
    62: "sys_ustat",
    63: "sys_dup2",
    64: "sys_getppid",
    65: "sys_getpgrp",
    66: "sys_setsid",
    67: "sys_sigaction",
    70: "sys_setreuid16",
    71: "sys_setregid16",
    72: "sys_sigsuspend",
    73: "sys_sigpending",
    74: "sys_sethostname",
    75: "sys_setrlimit",
    76: "sys_old_getrlimit",
    77: "sys_getrusage",
    78: "sys_gettimeofday",
    79: "sys_settimeofday",
    80: "sys_getgroups16",
    81: "sys_setgroups16",
    82: "sys_old_select",
    83: "sys_symlink",
    85: "sys_readlink",
    86: "sys_uselib",
    87: "sys_swapon",
    88: "sys_reboot",
    89: "sys_old_readdir",
    90: "sys_old_mmap",
    91: "sys_munmap",
    92: "sys_truncate",
    93: "sys_ftruncate",
    94: "sys_fchmod",
    95: "sys_fchown16",
    96: "sys_getpriority",
    97: "sys_setpriority",
    99: "sys_statfs",
    100: "sys_fstatfs",
    102: "sys_socketcall",
    103: "sys_syslog",
    104: "sys_setitimer",
    105: "sys_getitimer",
    106: "sys_newstat",
    107: "sys_newlstat",
    108: "sys_newfstat",
    111: "sys_vhangup",
    113: "sys_syscall",
    114: "sys_wait4",
    115: "sys_swapoff",
    116: "sys_sysinfo",
    117: "sys_ipc",
    118: "sys_fsync",
    119: "sys_sigreturn_wrapper",
    120: "sys_clone",
    121: "sys_setdomainname",
    122: "sys_newuname",
    124: "sys_adjtimex",
    125: "sys_mprotect",
    126: "sys_sigprocmask",
    128: "sys_init_module",
    129: "sys_delete_module",
    131: "sys_quotactl",
    132: "sys_getpgid",
    133: "sys_fchdir",
    134: "sys_bdflush",
    135: "sys_sysfs",
    136: "sys_personality",
    138: "sys_setfsuid16",
    139: "sys_setfsgid16",
    140: "sys_llseek",
    141: "sys_getdents",
    142: "sys_select",
    143: "sys_flock",
    144: "sys_msync",
    145: "sys_readv",
    146: "sys_writev",
    147: "sys_getsid",
    148: "sys_fdatasync",
    149: "sys_sysctl",
    150: "sys_mlock",
    151: "sys_munlock",
    152: "sys_mlockall",
    153: "sys_munlockall",
    154: "sys_sched_setparam",
    155: "sys_sched_getparam",
    156: "sys_sched_setscheduler",
    157: "sys_sched_getscheduler",
    158: "sys_sched_yield",
    159: "sys_sched_get_priority_max",
    160: "sys_sched_get_priority_min",
    161: "sys_sched_rr_get_interval",
    162: "sys_nanosleep",
    163: "sys_mremap",
    164: "sys_setresuid16",
    165: "sys_getresuid16",
    168: "sys_poll",
    170: "sys_setresgid16",
    171: "sys_getresgid16",
    172: "sys_prctl",
    173: "sys_rt_sigreturn_wrapper",
    174: "sys_rt_sigaction",
    175: "sys_rt_sigprocmask",
    176: "sys_rt_sigpending",
    177: "sys_rt_sigtimedwait",
    178: "sys_rt_sigqueueinfo",
    179: "sys_rt_sigsuspend",
    180: "sys_pread64",
    181: "sys_pwrite64",
    182: "sys_chown16",
    183: "sys_getcwd",
    184: "sys_capget",
    185: "sys_capset",
    186: "sys_sigaltstack",
    187: "sys_sendfile",
    190: "sys_vfork",
    191: "sys_getrlimit",
    192: "sys_mmap2",
    193: "sys_truncate64",
    194: "sys_ftruncate64",
    195: "sys_stat64",
    196: "sys_lstat64",
    197: "sys_fstat64",
    198: "sys_lchown",
    199: "sys_getuid",
    200: "sys_getgid",
    201: "sys_geteuid",
    202: "sys_getegid",
    203: "sys_setreuid",
    204: "sys_setregid",
    205: "sys_getgroups",
    206: "sys_setgroups",
    207: "sys_fchown",
    208: "sys_setresuid",
    209: "sys_getresuid",
    210: "sys_setresgid",
    211: "sys_getresgid",
    212: "sys_chown",
    213: "sys_setuid",
    214: "sys_setgid",
    215: "sys_setfsuid",
    216: "sys_setfsgid",
    217: "sys_getdents64",
    218: "sys_pivot_root",
    219: "sys_mincore",
    220: "sys_madvise",
    221: "sys_fcntl64",
    224: "sys_gettid",
    225: "sys_readahead",
    226: "sys_setxattr",
    227: "sys_lsetxattr",
    228: "sys_fsetxattr",
    229: "sys_getxattr",
    230: "sys_lgetxattr",
    231: "sys_fgetxattr",
    232: "sys_listxattr",
    233: "sys_llistxattr",
    234: "sys_flistxattr",
    235: "sys_removexattr",
    236: "sys_lremovexattr",
    237: "sys_fremovexattr",
    238: "sys_tkill",
    239: "sys_sendfile64",
    240: "sys_futex",
    241: "sys_sched_setaffinity",
    242: "sys_sched_getaffinity",
    243: "sys_io_setup",
    244: "sys_io_destroy",
    245: "sys_io_getevents",
    246: "sys_io_submit",
    247: "sys_io_cancel",
    248: "sys_exit_group",
    249: "sys_lookup_dcookie",
    250: "sys_epoll_create",
    251: "sys_epoll_ctl",
    252: "sys_epoll_wait",
    253: "sys_remap_file_pages",
    256: "sys_set_tid_address",
    257: "sys_timer_create",
    258: "sys_timer_settime",
    259: "sys_timer_gettime",
    260: "sys_timer_getoverrun",
    261: "sys_timer_delete",
    262: "sys_clock_settime",
    263: "sys_clock_gettime",
    264: "sys_clock_getres",
    265: "sys_clock_nanosleep",
    266: "sys_statfs64_wrapper",
    267: "sys_fstatfs64_wrapper",
    268: "sys_tgkill",
    269: "sys_utimes",
    270: "sys_arm_fadvise64_64",
    271: "sys_pciconfig_iobase",
    272: "sys_pciconfig_read",
    273: "sys_pciconfig_write",
    274: "sys_mq_open",
    275: "sys_mq_unlink",
    276: "sys_mq_timedsend",
    277: "sys_mq_timedreceive",
    278: "sys_mq_notify",
    279: "sys_mq_getsetattr",
    280: "sys_waitid",
    281: "sys_socket",
    282: "sys_bind",
    283: "sys_connect",
    284: "sys_listen",
    285: "sys_accept",
    286: "sys_getsockname",
    287: "sys_getpeername",
    288: "sys_socketpair",
    289: "sys_send",
    290: "sys_sendto",
    291: "sys_recv",
    292: "sys_recvfrom",
    293: "sys_shutdown",
    294: "sys_setsockopt",
    295: "sys_getsockopt",
    296: "sys_sendmsg",
    297: "sys_recvmsg",
    298: "sys_semop",
    299: "sys_semget",
    300: "sys_semctl",
    301: "sys_msgsnd",
    302: "sys_msgrcv",
    303: "sys_msgget",
    304: "sys_msgctl",
    305: "sys_shmat",
    306: "sys_shmdt",
    307: "sys_shmget",
    308: "sys_shmctl",
    309: "sys_add_key",
    310: "sys_request_key",
    311: "sys_keyctl",
    312: "sys_semtimedop",
    314: "sys_ioprio_set",
    315: "sys_ioprio_get",
    316: "sys_inotify_init",
    317: "sys_inotify_add_watch",
    318: "sys_inotify_rm_watch",
    319: "sys_mbind",
    320: "sys_get_mempolicy",
    321: "sys_set_mempolicy",
    322: "sys_openat",
    323: "sys_mkdirat",
    324: "sys_mknodat",
    325: "sys_fchownat",
    326: "sys_futimesat",
    327: "sys_fstatat64",
    328: "sys_unlinkat",
    329: "sys_renameat",
    330: "sys_linkat",
    331: "sys_symlinkat",
    332: "sys_readlinkat",
    333: "sys_fchmodat",
    334: "sys_faccessat",
    335: "sys_pselect6",
    336: "sys_ppoll",
    337: "sys_unshare",
    338: "sys_set_robust_list",
    339: "sys_get_robust_list",
    340: "sys_splice",
    341: "sys_sync_file_range2",
    342: "sys_tee",
    343: "sys_vmsplice",
    344: "sys_move_pages",
    345: "sys_getcpu",
    346: "sys_epoll_pwait",
    347: "sys_kexec_load",
    348: "sys_utimensat",
    349: "sys_signalfd",
    350: "sys_timerfd_create",
    351: "sys_eventfd",
    352: "sys_fallocate",
    353: "sys_timerfd_settime",
    354: "sys_timerfd_gettime",
    355: "sys_signalfd4",
    356: "sys_eventfd2",
    357: "sys_epoll_create1",
    358: "sys_dup3",
    359: "sys_pipe2",
    360: "sys_inotify_init1",
    361: "sys_preadv",
    362: "sys_pwritev",
    363: "sys_rt_tgsigqueueinfo",
    364: "sys_perf_event_open",
    365: "sys_recvmmsg",
    366: "sys_accept4",
    367: "sys_fanotify_init",
    368: "sys_fanotify_mark",
    369: "sys_prlimit64",
    370: "sys_name_to_handle_at",
    371: "sys_open_by_handle_at",
    372: "sys_clock_adjtime",
    373: "sys_syncfs",
    374: "sys_sendmmsg",
    375: "sys_setns",
    376: "sys_process_vm_readv",
    377: "sys_process_vm_writev",
    378: "sys_kcmp",
    379: "sys_finit_module",
    380: "sys_sched_setattr",
    381: "sys_sched_getattr",
    382: "sys_renameat2",
    383: "sys_seccomp",
    384: "sys_getrandom",
    385: "sys_memfd_create",
    386: "sys_bpf",
    387: "sys_execveat",
    388: "sys_userfaultfd",
    389: "sys_membarrier",
    390: "sys_mlock2",
    391: "sys_copy_file_range",
    392: "sys_preadv2",
    393: "sys_pwritev2",
    394: "sys_pkey_mprotect",
    395: "sys_pkey_alloc",
    396: "sys_pkey_free",
    397: "sys_statx",
    983041: "sys_ARM_NR_breakpoint",
    983042: "sys_ARM_NR_cacheflush",
    983043: "sys_ARM_NR_usr26",
    983044: "sys_ARM_NR_usr32",
    983045: "sys_ARM_NR_set_tls",
}
aarch64 = {
    0: "sys_io_setup",
    1: "sys_io_destroy",
    2: "sys_io_submit",
    3: "sys_io_cancel",
    4: "sys_io_getevents",
    5: "sys_setxattr",
    6: "sys_lsetxattr",
    7: "sys_fsetxattr",
    8: "sys_getxattr",
    9: "sys_lgetxattr",
    10: "sys_fgetxattr",
    11: "sys_listxattr",
    12: "sys_llistxattr",
    13: "sys_flistxattr",
    14: "sys_removexattr",
    15: "sys_lremovexattr",
    16: "sys_fremovexattr",
    17: "sys_getcwd",
    18: "sys_lookup_dcookie",
    19: "sys_eventfd2",
    20: "sys_epoll_create1",
    21: "sys_epoll_ctl",
    22: "sys_epoll_pwait",
    23: "sys_dup",
    24: "sys_dup3",
    25: "sys_fcntl",
    26: "sys_inotify_init1",
    27: "sys_inotify_add_watch",
    28: "sys_inotify_rm_watch",
    29: "sys_ioctl",
    30: "sys_ioprio_set",
    31: "sys_ioprio_get",
    32: "sys_flock",
    33: "sys_mknodat",
    34: "sys_mkdirat",
    35: "sys_unlinkat",
    36: "sys_symlinkat",
    37: "sys_linkat",
    38: "sys_renameat",
    39: "sys_umount",
    40: "sys_mount",
    41: "sys_pivot_root",
    43: "sys_statfs",
    44: "sys_fstatfs",
    45: "sys_truncate",
    46: "sys_ftruncate",
    47: "sys_fallocate",
    48: "sys_faccessat",
    49: "sys_chdir",
    50: "sys_fchdir",
    51: "sys_chroot",
    52: "sys_fchmod",
    53: "sys_fchmodat",
    54: "sys_fchownat",
    55: "sys_fchown",
    56: "sys_openat",
    57: "sys_close",
    58: "sys_vhangup",
    59: "sys_pipe2",
    60: "sys_quotactl",
    61: "sys_getdents64",
    62: "sys_lseek",
    63: "sys_read",
    64: "sys_write",
    65: "sys_readv",
    66: "sys_writev",
    67: "sys_pread64",
    68: "sys_pwrite64",
    69: "sys_preadv",
    70: "sys_pwritev",
    71: "sys_sendfile64",
    72: "sys_pselect6",
    73: "sys_ppoll",
    74: "sys_signalfd4",
    75: "sys_vmsplice",
    76: "sys_splice",
    77: "sys_tee",
    78: "sys_readlinkat",
    79: "sys_newfstatat",
    80: "sys_newfstat",
    81: "sys_sync",
    82: "sys_fsync",
    83: "sys_fdatasync",
    84: "sys_sync_file_range",
    85: "sys_timerfd_create",
    86: "sys_timerfd_settime",
    87: "sys_timerfd_gettime",
    88: "sys_utimensat",
    89: "sys_acct",
    90: "sys_capget",
    91: "sys_capset",
    92: "sys_personality",
    93: "sys_exit",
    94: "sys_exit_group",
    95: "sys_waitid",
    96: "sys_set_tid_address",
    97: "sys_unshare",
    98: "sys_futex",
    99: "sys_set_robust_list",
    100: "sys_get_robust_list",
    101: "sys_nanosleep",
    102: "sys_getitimer",
    103: "sys_setitimer",
    104: "sys_kexec_load",
    105: "sys_init_module",
    106: "sys_delete_module",
    107: "sys_timer_create",
    108: "sys_timer_gettime",
    109: "sys_timer_getoverrun",
    110: "sys_timer_settime",
    111: "sys_timer_delete",
    112: "sys_clock_settime",
    113: "sys_clock_gettime",
    114: "sys_clock_getres",
    115: "sys_clock_nanosleep",
    116: "sys_syslog",
    117: "sys_ptrace",
    118: "sys_sched_setparam",
    119: "sys_sched_setscheduler",
    120: "sys_sched_getscheduler",
    121: "sys_sched_getparam",
    122: "sys_sched_setaffinity",
    123: "sys_sched_getaffinity",
    124: "sys_sched_yield",
    125: "sys_sched_get_priority_max",
    126: "sys_sched_get_priority_min",
    127: "sys_sched_rr_get_interval",
    128: "sys_restart_syscall",
    129: "sys_kill",
    130: "sys_tkill",
    131: "sys_tgkill",
    132: "sys_sigaltstack",
    133: "sys_rt_sigsuspend",
    134: "sys_rt_sigaction",
    135: "sys_rt_sigprocmask",
    136: "sys_rt_sigpending",
    137: "sys_rt_sigtimedwait",
    138: "sys_rt_sigqueueinfo",
    139: "sys_rt_sigreturn",
    140: "sys_setpriority",
    141: "sys_getpriority",
    142: "sys_reboot",
    143: "sys_setregid",
    144: "sys_setgid",
    145: "sys_setreuid",
    146: "sys_setuid",
    147: "sys_setresuid",
    148: "sys_getresuid",
    149: "sys_setresgid",
    150: "sys_getresgid",
    151: "sys_setfsuid",
    152: "sys_setfsgid",
    153: "sys_times",
    154: "sys_setpgid",
    155: "sys_getpgid",
    156: "sys_getsid",
    157: "sys_setsid",
    158: "sys_getgroups",
    159: "sys_setgroups",
    160: "sys_newuname",
    161: "sys_sethostname",
    162: "sys_setdomainname",
    163: "sys_getrlimit",
    164: "sys_setrlimit",
    165: "sys_getrusage",
    166: "sys_umask",
    167: "sys_prctl",
    168: "sys_getcpu",
    169: "sys_gettimeofday",
    170: "sys_settimeofday",
    171: "sys_adjtimex",
    172: "sys_getpid",
    173: "sys_getppid",
    174: "sys_getuid",
    175: "sys_geteuid",
    176: "sys_getgid",
    177: "sys_getegid",
    178: "sys_gettid",
    179: "sys_sysinfo",
    180: "sys_mq_open",
    181: "sys_mq_unlink",
    182: "sys_mq_timedsend",
    183: "sys_mq_timedreceive",
    184: "sys_mq_notify",
    185: "sys_mq_getsetattr",
    186: "sys_msgget",
    187: "sys_msgctl",
    188: "sys_msgrcv",
    189: "sys_msgsnd",
    190: "sys_semget",
    191: "sys_semctl",
    192: "sys_semtimedop",
    193: "sys_semop",
    194: "sys_shmget",
    195: "sys_shmctl",
    196: "sys_shmat",
    197: "sys_shmdt",
    198: "sys_socket",
    199: "sys_socketpair",
    200: "sys_bind",
    201: "sys_listen",
    202: "sys_accept",
    203: "sys_connect",
    204: "sys_getsockname",
    205: "sys_getpeername",
    206: "sys_sendto",
    207: "sys_recvfrom",
    208: "sys_setsockopt",
    209: "sys_getsockopt",
    210: "sys_shutdown",
    211: "sys_sendmsg",
    212: "sys_recvmsg",
    213: "sys_readahead",
    214: "sys_brk",
    215: "sys_munmap",
    216: "sys_mremap",
    217: "sys_add_key",
    218: "sys_request_key",
    219: "sys_keyctl",
    220: "sys_clone",
    221: "sys_execve",
    222: "sys_mmap",
    223: "sys_fadvise64_64",
    224: "sys_swapon",
    225: "sys_swapoff",
    226: "sys_mprotect",
    227: "sys_msync",
    228: "sys_mlock",
    229: "sys_munlock",
    230: "sys_mlockall",
    231: "sys_munlockall",
    232: "sys_mincore",
    233: "sys_madvise",
    234: "sys_remap_file_pages",
    235: "sys_mbind",
    236: "sys_get_mempolicy",
    237: "sys_set_mempolicy",
    238: "sys_migrate_pages",
    239: "sys_move_pages",
    240: "sys_rt_tgsigqueueinfo",
    241: "sys_perf_event_open",
    242: "sys_accept4",
    243: "sys_recvmmsg",
    260: "sys_wait4",
    261: "sys_prlimit64",
    262: "sys_fanotify_init",
    263: "sys_fanotify_mark",
    264: "sys_name_to_handle_at",
    265: "sys_open_by_handle_at",
    266: "sys_clock_adjtime",
    267: "sys_syncfs",
    268: "sys_setns",
    269: "sys_sendmmsg",
    270: "sys_process_vm_readv",
    271: "sys_process_vm_writev",
    272: "sys_kcmp",
    273: "sys_finit_module",
    274: "sys_sched_setattr",
    275: "sys_sched_getattr",
    276: "sys_renameat2",
    277: "sys_seccomp",
    278: "sys_getrandom",
    279: "sys_memfd_create",
    280: "sys_bpf",
    281: "sys_execveat",
    282: "sys_userfaultfd",
    283: "sys_membarrier",
    284: "sys_mlock2",
    285: "sys_copy_file_range",
    286: "sys_preadv2",
    287: "sys_pwritev2",
    288: "sys_pkey_mprotect",
    289: "sys_pkey_alloc",
    290: "sys_pkey_free",
    291: "sys_statx",
}
